"""Methods to save and/or reload libcbm output"""

from functools import cached_property
from eu_cbm_hat.info.output_data import OutputData

from autopaths.auto_paths import AutoPaths

class BudSim:
    """Dummy class to attach sit at a sub level similar to runner.simulation.sit
    """

    def __init__(self, parent):
        self.bud = parent

    @property
    def sit(self):
        """SIT"""
        return self.bud.sit

    @property
    def cbm_output(self):
        """SIT"""
        return self.bud.cbm_output

class BudOutput(OutputData):
    """libcbm simulation output"""

    def __init__(self, parent):
        self.parent = parent
        self.runner = parent
        self.tables = ['area', 'classifiers', 'flux', 'parameters', 'pools', 'state']
        # Properties defined to be able to reuse runner methods
        # TODO: create a self.sim such that so that we can use methods of
        # info/output_data.py which inherits from info/internal_data.py
        # self.sim.sit = self.bud.sit
        self.paths = AutoPaths(str(self.parent.data_dir), self.all_paths)
        self.sim = self.parent.sim

    # def save(self):
    #     """Save libcbm output to parquet files on disk

    #     One parquet file for each of the tables:
    #     ['area', 'classifiers', 'flux', 'parameters', 'pools', 'state']

    #     """

    #     for t in self.tables:
    #         print(self.bud.cbm_output[t].head())




""" The purpose of this script is to compare libcbm 1x and libcbm 2x

Warning: This script will only work on a development machine which has libcbm_py
and eu_cbm_hat installed from git repositories and given as a path through the
PYTHONPATH environment variable.

- Results are written to a parquet file in eu_cbm_hat/info/output_data.py
- This result dataframe is a merge of all results data frames from the cbm_output object
- These data frames are read in the `__getitem__` method of the class InternalData
- That `__getitem__` method calls `getattr(self.sim.cbm_output, item).to_pandas()`

For the selected countries and for the 2 different versions of the model,
backup these parquet files in a "comparison" directory inside the eu_cbm_data
directory.

"""

import shutil
import sys
import warnings

from pathlib import Path
import pandas
from eu_cbm_hat.core.continent import continent
import git


# Choose which libcbm version to run
RUN_LIBCBM_VERSION = None
# RUN_LIBCBM_VERSION = 1
# RUN_LIBCBM_VERSION = 2

# Destination directory to store and compare results
comp_dir = Path(continent.base_dir) / "output" / "comparison"
if not comp_dir.exists():
    comp_dir.mkdir()

# Find location of git repositories for libcbm_py and eu_cbm_hat
def find_sys_path(path_contains:str):
    """Find path that contains the given characters.
    Raise an error if there's not exactly one matching path"""
    matching_paths = [path for path in sys.path if path_contains in path]
    if len(matching_paths) != 1:
        msg = f"Expected one path containing {path_contains}, "
        msg += f"found {len(matching_paths)}\n"
        msg += f"{matching_paths}"
        raise ValueError(msg)
    return matching_paths[0]

repo_eu_cbm_hat = git.Repo(find_sys_path("eu_cbm_hat"))
repo_libcbm_py = git.Repo(find_sys_path("libcbm_py"))

def checkout_branch(git_repo:git.repo.base.Repo, branch_name:str):
    """Check if a repository has any changes and checkout the given branch
    """
    if git_repo.is_dirty(untracked_files=True):
        msg = f"There are changes in {git_repo}.\n"
        msg += f"Not checking out the '{branch_name}' branch."
        raise RuntimeError(msg)
    git_repo.git.checkout(branch_name)
    print(f"Checked out branch: {branch_name} of {git_repo}.")


# Create runners
r_at = continent.combos['reference'].runners['AT'][-1]
r_cz = continent.combos['reference'].runners['CZ'][-1]
r_se = continent.combos['reference'].runners['SE'][-1]
r_zz = continent.combos['reference'].runners['ZZ'][-1]

def run_and_store_comparison(runner, version):
    """Run the model and store results in parquet files for comparison purposes."""
    print(f"Running {runner}\nwith libcbm version {version}")
    runner.num_timesteps = 30
    runner.run(keep_in_ram=True, verbose=True, interrupt_on_error=True)
    new_parquet_file = comp_dir / f"zz_output_libcbm_{version}.parquet"
    shutil.copy(runner.output.paths["results"], new_parquet_file)


######################################################################
# Run Libcbm version 1 and store results in the comparison directory #
######################################################################
if RUN_LIBCBM_VERSION == 1:
    warnings.warn("You need to restart the python shell between changes of libcbm version.")
    checkout_branch(repo_libcbm_py, "1.x")
    checkout_branch(repo_eu_cbm_hat, "main")
    assert repo_libcbm_py.active_branch.name == '1.x'
    assert repo_eu_cbm_hat.active_branch.name == 'main'

    # Old comment which may not be relevant now that we have the
    # checkout_branch function to checkout the required branches.
    # eu_cbm_hat v0.6.1 is compatible with libcbm version 1
    # Checkout the libcbm 1x branch and tag v0.6.1 of eu_cbm_hat
    # cd ~/repos/eu_cbm/libcbm_py/
    # git checkout 1.x
    # cd ~/repos/eu_cbm/eu_cbm_hat/
    # git checkout tags/v0.6.1

    # Run the models
    for this_runner in [r_zz]:
        run_and_store_comparison(this_runner, RUN_LIBCBM_VERSION)

######################################################################
# Run Libcbm version 2 and store results in the comparison directory #
######################################################################
if RUN_LIBCBM_VERSION == 2:
    warnings.warn("You need to restart the python shell between changes of libcbm version.")
    checkout_branch(repo_libcbm_py, "2.x")
    checkout_branch(repo_eu_cbm_hat, "libcbm2")
    assert repo_libcbm_py.active_branch.name == '2.x'
    assert repo_eu_cbm_hat.active_branch.name == 'libcbm2'
    # Run the models
    for this_runner in [r_zz]:
        run_and_store_comparison(this_runner, RUN_LIBCBM_VERSION)

###################
# Compare results #
###################
# Compare the resulting parquet files
zz1 = pandas.read_parquet(comp_dir / "zz_output_libcbm_1.parquet")
zz2 = pandas.read_parquet(comp_dir / "zz_output_libcbm_2.parquet")


#################################
# Dumb comparison line for line #
#################################
# Check the ones with different disturbance types
zz3 = zz2.copy()
zz3["diff"] = zz1["disturbance_type"] - zz2["disturbance_type"]
zz3["disturbance_type_v1"] = zz1["disturbance_type"]
zz3["disturbance_type_v2"] = zz2["disturbance_type"]
zz3.query("diff!=0")
# It seems the data frames are not aligned

index = ['timestep',
         'disturbance_type',
         'status',
         'forest_type',
         'region',
         'mgmt_type',
         'mgmt_strategy',
         'climate',
         'con_broad',
         'site_index',
         'growth_period',
         'year',
         'age_class',
        ]

# Sort values by the index
zz1.sort_values(index, inplace=True)
zz1.reset_index(inplace=True, drop=True)
zz2.sort_values(index, inplace=True)
zz2.reset_index(inplace=True, drop=True)

# There are less differences now
zz4 = zz2.copy()
zz4["diff"] = zz1["disturbance_type"] - zz2["disturbance_type"]
zz4["disturbance_type_v1"] = zz1["disturbance_type"]
zz4["disturbance_type_v2"] = zz2["disturbance_type"]
zz4.query("diff!=0")
# zz4.query("diff!=0").to_csv("/tmp/zz4.csv")






print("zz1.equals(zz2):", zz1.equals(zz2))

for col in zz1.columns:
    if pandas.api.types.is_numeric_dtype(zz1[col]):
        diff = zz1[col] - zz2[col]
        print(f"\n{col}:", diff.abs().sum() / zz1[col].sum())
        print(diff)
    else:
        print(f"{col} is of string type")


#########################
# Aggregate and compare #
#########################
# Aggregate area and merchantable pool over
# only a few classifiers and disturbance type.
# Then compare version 1 and 2.
index = ["year", "forest_type", "region", "disturbance_type"]
agg_vars = ["area",
            "hardwood_merch",
            "softwood_merch",
            "softwood_merch_to_product",
            "hardwood_merch_to_product",
            ]
zz1_agg = zz1.groupby(index)[agg_vars].agg(sum).reset_index()
zz2_agg = zz2.groupby(index)[agg_vars].agg(sum).reset_index()

# Merge
zz5_agg = zz1_agg.merge(zz2_agg, on=index, how="outer")

# diff
zz5_agg["hwm_diff"]  = zz5_agg["hardwood_merch_y"] - zz5_agg["hardwood_merch_x"]
zz5_agg["swm_diff"]  = zz5_agg["softwood_merch_y"] - zz5_agg["softwood_merch_x"]
# zz5_agg.query("hwm_diff !=0 or swm_diff !=0").to_csv("/tmp/hwmdiff.csv", index=False)






####################################################################
# Make a reproducible data set that can be run with libcbm_py only #
####################################################################






""" Model runs for the Carbon Removal and Carbon Farming project

A subpackage that reuses some of the EU-CBM-HAT classes to build a dedicated
libcbm model runner.

"""
from eu_cbm_hat.crcf.runner import Runner
from eu_cbm_hat.crcf.country_simulator import CountrySimulator


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair, Paul Rougieux and Viorel Blujdea.

JRC Biomass Project.
Unit D1 Bioeconomy.


# Introduction

This documentation describes EU-CBM-HAT an adaptation of the Carbon Budget
Model (CBM) for EU countries, with the addition of a Harvest Allocation Tool
(HAT). `eu_cbm_hat` is a python package that enables the assessment of forest
CO2 emissions and removals under scenarios of forest management, natural
disturbances and forest-related land use changes. For more information on the
structure of the model and the research questions that have been analysed see
the section on related publication below.

EU-CBM-HAT depends on a software library called
[libcbm](https://github.com/cat-cfs/libcbm_py). libcbm was developed and is
maintained by the Forest carbon accounting team of Natural Resources Canada.
The `eu_cbm_hat` package was developed on top of this library by the Joint
Research Centre (JRC) of the European Commission to add capabilities such
as: scenario processing, a data structure for EU counties, a harvest allocation
tool, growth modification and estimates of the harvested wood products sink.
eu_cbm_hat is publicly available on the python package index at
[https://pypi.org/project/eu-cbm-hat](https://pypi.org/project/eu-cbm-hat). The
source code is available openly on
[https://gitlab.com/bioeconomy/eu_cbm/eu_cbm_hat](https://gitlab.com/bioeconomy/eu_cbm/eu_cbm_hat)
and the documentation you are currently reading is available at
[https://bioeconomy.gitlab.io/eu_cbm/eu_cbm_hat/eu_cbm_hat.html](https://bioeconomy.gitlab.io/eu_cbm/eu_cbm_hat/eu_cbm_hat.html).


There are 3 ways to run the model:

1. Using a [runner](eu_cbm_hat/core/runner.html) created from simulation tools
  in the [core](eu_cbm_hat/core.html) directory. This runner has been developed
  since 2019. It requires a comprehensive data structure for EU countries to be
  defined in an eu_cbm_data directory. An example of the directory structure is
  visible in the test subdirectory for a fictitious country called ZZ.

2. Using a smaller runner called a [bud]() object. This approach has been added
  in 2025. A bud is a  much simpler object that feeds a simpler data directory to
  libcbm. It can use the same post-processor methods as the runner.

3. Using another runner called crcf made for the simulations of Carbon Removal and
  Carbon Farming. Introduced in 2025, this is not documented yet. See the source
  code in the crcf directory.


# Common methods used by all runner types

- Paths to data directories are defined in
  [constants](eu_cbm_hat/constants.html). Other variables necessary for
  computations such as the carbon fraction of biomass are also defined there.

- During the libcbm simulation time step, various modifications can be made to
  the growth or disturbances:

    - The Harvest Allocation Tool implemented in [cbm.dynamic](eu_cbm_hat/cbm/dynamic.html)
      provides the capability for dynamic disturbance allocation depending on the
      evolution of the stock. It also deals with salvage logging after natural
      disturbances.

    - [cbm.climate_growth_modifier](eu_cbm_hat/cbm/climate_growth_modifier.html)
      can modify forest growth at each time step in order to simulate the impact
      of climate variables such as draught on forest growth. The input data for
      this is derived from NPP measures or simulations.

- The [post_processor](eu_cbm_hat/post_processor.html) transforms CBM output
  fluxes and pools tables into final result tables.

    - [sink](eu_cbm_hat/post_processor/sink.html) computes the carbon sink in
      tons of CO2 equivalent

    - [stock](eu_cbm_hat/post_processor/stock.html) computes stock indicators

    - [hwp](eu_cbm_hat/post_processor/hwp.html) estimates Harvested Wood
      Products inflows and outflows to compute the HWP sink.


# Runner methods

- Scenario combinations are defined as `.yaml` files in `eu_cbm_data/combos`.
  When initiating a runner, the content of its yaml file is loaded in
  [eu_cbm_hat.combos](eu_cbm_hat/combos.html). It defines the scenario used for
  all input files. Some input files can change scenario every year, other files
  are defined for the entirety of the simulation length.

- A runner processes data in several steps.

    - [orig_data](eu_cbm_hat/info/orig_data.html) contains the original data
      for a country. There are many scenarios for inventory, growth (yield), and
      disturbances.

    - [aidb](eu_cbm_hat/info/aidb.html) contains the Archive Index Database
      (with soil parameters and biomass expansion factors)

    - [input_data](eu_cbm_hat/info/input_data.html) contains the actual input
      data sent to libcbm for one and only one combination of scenarios.

    - [output_data](eu_cbm_hat/info/output_data.html) contains the output
      fluxes and pools after the libcbm simulation run.

    - The ouput data is then used by the
      [post_processor](eu_cbm_hat/post_processor.html) see below.


# Bud methods

[Bud]() is a small self contain runner-type object to run libcbm by pointing it to an input
data directory and an AIDB. It is a small self contained object that makes it
possible to run the libcbm model and the EU-CBM-HAT post processor (to compute
sink output for example) without the need for the EU-wide
eu_cbm_data directory.

The data is processed by sub modules:

- [input_data](eu_cbm_hat/bud/input_data.html) prepares the input data for libcbm.

- [output](eu_cbm_hat/bud/output.html) handles the output data.

- The output data is further processed in the [bud
  post_processor](eu_cbm_hat/bud/post_processor.html) which inherits all methods
  from the main [post_processor](eu_cbm_hat/post_processor.html), for example
  methods to compute the stock, sink and hwp mentioned in the common methods
  section above.


# CRCF methods

Runner made for simulations of Carbon Removal and Carbon Farming. This type of
object is not documented yet. Look at the source code in the crcf directory.


# Related Publications

In chronological order.

- Kurz WA, Dymond CC, White TM, Stinson G, Shaw CH, Rampley GJ, Smyth C,
  Simpson BN, Neilson ET, Trofymow JA, Metsaranta J. **CBM-CFS3: a model of
  carbon-dynamics in forestry and land-use change implementing IPCC standards**.
  Ecological modelling. 2009 Feb 24;220(4):480-504.
  https://doi.org/10.1016/j.ecolmodel.2008.10.018

  - This is the main publication describing the Carbon Budget Model structure.
    It is a previous version called CBM-CFS3.


- Roberto Pilli, Giacomo Grassi, Werner A. Kurz, Carolyn E. Smyth, Viorel
  Blujdea, **Application of the CBM-CFS3 model to estimate Italy's forest carbon
  budget, 1995–2020,** Ecological Modelling, Volume 266, 2013, Pages 144-171,
  ISSN 0304-3800, https://doi.org/10.1016/j.ecolmodel.2013.07.007.

    - This paper used a previous version called CBM-CFS3. Figure 3 in the paper
      explains the need to switch from an init (historical) growth period to a
      current growth period. The
      [Simulation.switch_period](eu_cbm_hat/cbm/simulation.html#Simulation.switch_period)
      method changes the growth period from "Init" to "Cur". As a result all
      stands will have the value "Cur" for the growth_period classifier from
      inventory start year onwards. Note the inventory start year is stored in
      `runner.country.inventory_start_year` and defined in
      [Country.set_years](eu_cbm_hat/core/country.html#Country.set_years).

- Pilli, Roberto, Blujdea, Viorel N.B., Rougieux, Paul, Grassi, Giacomo,
  Mubareka, Sarah Betoul, 2024 Volume, increment, and aboveground biomass data
  series and biomass conversion and expansion factors for the main forest types
  of EU Member States Annals of Forest Science This collection reports the
  standing stock volume, increment, aboveground biomass, and biomass conversion
  and expansion factors attributed to 222 forest types and 48 different
  management types, representative of 25 EU Member States. Dataset
  https://zenodo.org/records/11387301. Article
  https://doi.org/10.1186/s13595-024-01256-5

- Blujdea, V., Rougieux, P., Sinclair, L., Morken, S., Pilli, R., Grassi, G.,
  Mubareka, S. and Kurz, W., **The JRC Forest Carbon Model: description of
  EU-CBM-HAT**, EUR 31299 EN, Publications Office of the European Union,
  Luxembourg, 2022, ISBN 978-92-76-58867-2, doi:10.2760/244051, JRC130609.
  https://publications.jrc.ec.europa.eu/repository/handle/JRC130609

- Pilli, R., Blujdea, V.N., Rougieux, P., Grassi, G. and Mubareka, S.B., **The
  calibration of the JRC EU Forest Carbon Model within the historical period 2010 - 2020**,
  Publications Office of the European Union, Luxembourg, 2024,
  doi:10.2760/222407, JRC135639. https://dx.doi.org/10.2760/222407

- Rougieux, P., Pilli, R., Blujdea, V., Mansuy, N. and Mubareka, S.B.,
  **Simulating future wood consumption and the impacts on Europe`s forest sink to
  2070**, Publications Office of the European Union, Luxembourg, 2024,
  doi:10.2760/17191, JRC136526.
  https://publications.jrc.ec.europa.eu/repository/handle/JRC136526


"""

# Special variables
__version__ = "2.1.2"

# Import constants first (no circular dependency)
from eu_cbm_hat.constants import (
    project_name,
    project_url,
    CARBON_FRACTION_OF_BIOMASS,
    module_dir,
    module_dir_pathlib,
    repos_dir,
    git_repo,
    eu_cbm_data_dir,
    eu_cbm_data_pathlib,
    eu_cbm_aidb_dir,
    eu_cbm_aidb_pathlib,
)


# Lazy import mechanism - only imports when actually accessed
def __getattr__(name):
    if name == "Bud":
        from eu_cbm_hat.bud import Bud

        return Bud
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


__all__ = [
    "bud",
    "cbm",
    "combos",
    "constants",
    "core",
    # Don't include "crcf" because it depends on `eu_cbm_crl`
    # which is not available.
    "info",
    "launch",
    "plot",
    "post_processor",
    "pump",
    "qaqc",
    "tests",
]



"""Update the AIDB from libcbm version 1 to libcbm version 2

Use the path to the AIDB repo defined in `eu_cbm_aidb_dir` so that we modify
those files directly.


Usage convert the AIDB from V1 to V2 in all countries:

    # Checkout branch "2.x" of cbm_defaults
    cd ~/eu_cbm/cbm_defaults/
    git checkout 2.x
    # Convert the AIDBs from version 1 to version 2
    ipython -i ~/eu_cbm/eu_cbm_hat/scripts/conversion/libcbmv2/update_aidb_1x_to_2x.py

Usage on one country only for debugging purposes to reproduce errors in a
specific country:

    >>> import sys
    >>> import pathlib
    >>> sys.path.append(str(pathlib.Path.home() /  "eu_cbm/eu_cbm_hat/scripts/conversion/libcbmv2/"))
    >>> from update_aidb_1x_to_2x import convert_aidb_to_v2
    >>> from update_aidb_1x_to_2x import countries_dir
    >>> convert_aidb_to_v2(countries_dir / "SI/aidb.db")
    >>> convert_aidb_to_v2(countries_dir / "IE/aidb.db")
    >>> convert_aidb_to_v2(countries_dir / "PL/aidb.db")

Display issues with unique constraints

    >>> from update_aidb_1x_to_2x import display_unique_constraint_issue
    >>> display_unique_constraint_issue(db_path = countries_dir / "PL/aidb.db",
    ...                                 table = "disturbance_matrix_value",
    ...                                 columns = ["disturbance_matrix_id",
    ...                                            "source_pool_id",
    ...                                            "sink_pool_id"])

See also:

    - The documentation on how to migrate from libcbm version 1 to libcbm
    version 2
    https://github.com/cat-cfs/cbm_defaults#migrating-database-version

    - An issue encountered when trying to use libcbm 2
    https://github.com/cat-cfs/libcbm_py/issues/58


"""

import pathlib
import shutil
import pandas
from cbm_defaults.update import db_updater
from eu_cbm_hat import eu_cbm_aidb_dir


from plumbing.databases.sqlite_database import SQLiteDatabase


countries_dir = pathlib.Path(eu_cbm_aidb_dir) / "countries"


def convert_aidb_to_v2(db_path):
    """Convert the aidb at the given path from V1 to V2"""
    if not db_path.exists():
        raise ValueError(f"There is no AIDB at {db_path}")
    # If it's already a version 2 table, do nothing
    df = pandas.read_sql_table("land_class", "sqlite:///" + str(db_path))
    if "land_type_id_1" in df.columns:
        msg = f"\n\n{db_path} is already a version 2 table because "
        msg += "the land_class table contains the land_type_id_1 column. "
        msg += "skip conversion."
        print(msg)
        return
    db_path_v1 = db_path.parent / "aidb_v1.db"
    db_path_v2 = db_path.parent / "aidb_v2.db"
    # Rename the old AIDB to v1
    shutil.copy(db_path, db_path_v1)
    # Remove the "index" column in the disturbance_matrix_value table.
    # This was introduced by mistake and shouldn't be there.
    db_v1 = SQLiteDatabase(str(db_path_v1))
    dm_values = db_v1.read_df("disturbance_matrix_value")
    dm_values = pandas.read_sql_table(
        "disturbance_matrix_value", "sqlite:///" + str(db_path_v1)
    )
    if "index" in dm_values.columns:
        dm_values.drop(columns="index", inplace=True)
        dm_values.to_sql(
            "disturbance_matrix_value",
            con="sqlite:///" + str(db_path_v1),
            index=False,
            if_exists="replace",
        )
    # Remove the V2 DB if already present
    if db_path_v2.exists():
        db_path_v2.unlink()
    # Update the AIDB from V1 to V2
    db_updater.update("1x_to_2x", db_path_v1, db_path_v2)
    # Make v2 the main table
    shutil.copy(db_path_v2, db_path)
    print(f"\n{db_path} updated to V2\n\n")

def display_unique_constraint_issue(db_path, table, columns):
    """Return rows that don't respect the unique constraint in the given table and column

    Example use:
        
        >>> # See above on how to add this script to your path if needed
        >>> from update_aidb_1x_to_2x import display_unique_constraint_issue
        >>> from update_aidb_1x_to_2x import countries_dir
        >>> db_path = countries_dir / "PL/aidb.db"
        >>> table = "disturbance_matrix_value"
        >>> columns = ["disturbance_matrix_id", "source_pool_id", "sink_pool_id"]
        >>> display_unique_constraint_issue(db_path, table, columns)

    """
    db = SQLiteDatabase(str(db_path))
    df = db.read_df(table)
    df[columns].duplicated()
    dup = df[columns].duplicated(keep=False)
    return df[dup]


if __name__ == "__main__":
    aidbs = countries_dir.glob("**/aidb.db")
    for this_db in aidbs:
        # this_db = countries_dir / "ZZ" / "aidb.db"
        try:
            convert_aidb_to_v2(this_db)
        # Catch all errors. This can be restricted to fewer errors.
        # Generally observed errors are:
        # - sqlite3.OperationalError table disturbance_matrix_value has no
        #   column named index
        # - sqlite3.IntegrityError UNIQUE constraint failed:
        #   vol_to_bio_factor.id
        except Exception as e:
            print(f"Error {this_db}")
            print(e)
            print("\n\n")



"""Copy test input to a temporary directory

"""

import tempfile
import shutil
from pathlib import Path
from eu_cbm_hat.constants import module_dir_pathlib

DATA_DIR = module_dir_pathlib / "tests/bud_data"

def copy_input_to_temp_dir() -> Path:
    """
    Copies the input data directory to a new temporary directory.
    Returns the Path to the temporary directory.
    """
    tmpdir = Path(tempfile.mkdtemp())
    temp_data_dir = tmpdir / DATA_DIR.name
    shutil.copytree(DATA_DIR, temp_data_dir)
    print(f"Temporary directory created at: {temp_data_dir}")
    return temp_data_dir




"""Climate adjustment variables based on modelled NPP values

Written by Viorel Blujdea and Paul Rougieux.

JRC Biomass Project. Unit D1 Bioeconomy.

The growth multiplier can have different meanings:

combined_multiplier = growth_multiplier_disturbance  X  climate_adjustement

| combined_multiplier | growth_multiplier_disturbance | climate_adjustement |
|-------------------  |-------------------------------|---------------------|
| =1.1 x 0.6          | 1.1                           | 0.6                 |
| =1.05 x 0.7         | 1.05                          | 0.7                 |
| =1.1 x 0.8          | 1.1                           | 0.8                 |
| =1 x 0.9            | 1                             | 0.9                 |
| =1 x 0.95           | 1                             | 0.95                |


According to libcbm_c source code's internal processing what is called
climate_adjustement is the growth_multiplier below:

    sw_multiplier = growthMult.SoftwoodMultiplier * growth_multiplier

"""

from functools import cached_property
from eu_cbm_hat.info.clim_adjust_common_input import (
    mean_npp_by_model_country_clu_con_broad,
)


class ClimAdjust:
    """Climate adjustment variables based on modelled NPP values

    >>> from eu_cbm_hat.core.continent import continent
    >>> runner = continent.combos['reference_cable_pop'].runners['EE'][-1]
    >>> # All model inputs for the given country
    >>> runner.clim_adjust.df_all

    >>> # Model input for the selected scenario and model as defined in the
    >>> # combo yaml file
    >>> runner.clim_adjust.df

    This data frame is used by cbm/climate_growth_modifier.py to feed
    growth multiplier to cbm within the time step.

    """

    def __init__(self, parent):
        self.runner = parent
        self.combo_name = self.runner.combo.short_name
        self.combo_config = self.runner.combo.config
        if "climate_adjustment_model" not in self.combo_config.keys():
            self.model = "default"
        else:
            self.model = self.combo_config["climate_adjustment_model"]
        if "climate_adjustment_hist_start_year" not in self.combo_config.keys():
            self.hist_start_year = None
        else:
            self.hist_start_year = self.combo_config[
                "climate_adjustment_hist_start_year"
            ]
        if "climate_adjustment_hist_end_year" not in self.combo_config.keys():
            self.hist_end_year = None
        else:
            self.hist_end_year = self.combo_config["climate_adjustment_hist_end_year"]

    @cached_property
    def df_all(self):
        """NPP values in all climate models for the given country"""
        country_name = self.runner.country.country_name
        df = mean_npp_by_model_country_clu_con_broad(
            hist_start_year=self.hist_start_year, hist_end_year=self.hist_end_year
        )
        selector = df["country"] == country_name
        return df.loc[selector].copy()

    @cached_property
    def df(self):
        """Climate model NPP inputs for the selected model in the given country

        Ignore the upper-case or lower-case in the model name selection.
        """
        df = self.df_all
        # Select the model, ignore the case
        selector = df["model"].str.lower() == self.model.lower()
        # Keep only those column
        cols = ["model", "country", "con_broad", "climate", "year", "npp", "ratio"]
        return df.loc[selector, cols].copy()



import numpy as np
from netCDF4 import Dataset



class Netcdf:

    def __init__(self, file_path:str):
        self.file = file_path
        self.is_init = False


    def init(self, num_timesteps:int, columns:list[str], cmp:str='zlib'):

        self.is_init = True

        # max length column labels
        nchrs = 0
        for label in columns:
            nchrs = max(nchrs, len(label))

        # init netcdf
        with Dataset(self.file, 'w') as nc:

            nc.createDimension('timestep', num_timesteps)
            nc.createDimension('column', len(columns))
            nc.createDimension('simulation', None)
            nc.createDimension('nchrs', nchrs)

            vlab = nc.createVariable('column_labels', 'S1', ('column', 'nchrs'), compression=cmp)
            vlab._Encoding = 'ascii'
            vlab[:] = np.array(columns, dtype=f'S{nchrs}')

            vsim = nc.createVariable('sim', 'f8', ('simulation', 'timestep', 'column'), compression=cmp)
            vsim.full_name = 'eu-cbm-hat simulations'


    def write_simulation(self, data:np.ndarray):
        with Dataset(self.file, 'a') as nc:
            idx = nc['sim'].shape[0]
            nc['sim'][idx] = data


"""Run the model in the Luigi workflow orchestration tool

Run the model in Luigi:

    cd /home/paul/eu_cbm/eu_cbm_hat/scripts/running
    luigi --module run_in_luigi RunCBM --combo reference --country-iso2 LU --final-year 2047 --local-scheduler

To re-run an already executed task, simply delete the output file.

Open the model output for exploration in python:

    >>> from eu_cbm_hat.core.continent import continent
    >>> runner = continent.combos['reference'].runners["LU"][-1]
    >>> runner.post_processor.sink

Display the luigi output file

    cat  ~/eu_cbm/eu_cbm_data/output/luigi/reference_LU_2047.txt

"""

from datetime import datetime
import luigi
from eu_cbm_hat.core.continent import continent
from eu_cbm_hat import eu_cbm_data_pathlib


class RunCBM(luigi.Task):
    """Run EU-CBM-HAT as a Luigi task"""
    combo = luigi.Parameter()
    country_iso2 = luigi.Parameter()
    final_year = luigi.IntParameter()

    def output(self):
        """Build the output path relative to eu_cbm_data"""
        output_dir = eu_cbm_data_pathlib / "output/luigi"
        output_file = output_dir / f"{self.combo}_{self.country_iso2}_{self.final_year}.txt"
        return luigi.LocalTarget(str(output_file))

    def run(self):
        """Run EU-CBM-HAT for the given country, scenario and final year

        Write the runner log file to the luigi output file."""
        runner = continent.combos[self.combo].runners[self.country_iso2][-1]
        runner.num_timesteps = self.final_year - runner.country.inventory_start_year
        runner.run(keep_in_ram=False, verbose=False, interrupt_on_error=False)
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        with self.output().open('w') as f:
            f.write(runner.paths.log.contents)
            msg = f"{now} Scenario combination '{self.combo}' model run completed "
            msg += f"for country '{self.country_iso2}' until {self.final_year}.\n"
            f.write(msg)

if __name__ == "__main__":
    luigi.run()



import pandas

from sqlalchemy import Column, ForeignKey, Integer, String, create_engine
from sqlalchemy.orm import DeclarativeBase, relationship, sessionmaker


class Base(DeclarativeBase):
    pass

class Status(Base):
    __tablename__ = 'status'
    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(10), unique=True)
    description = Column(String(100))
    species = relationship("Species", back_populates="status")

class ForestType(Base):
    __tablename__ = 'forest_type'
    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(10), unique=True)
    description = Column(String(100))
    species = relationship("Species", back_populates="forest_type")

class Region(Base):
    __tablename__ = 'region'
    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(10), unique=True)
    description = Column(String(100))
    species = relationship("Species", back_populates="region")

class ManagementType(Base):
    __tablename__ = 'mgmt_type'
    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(10), unique=True)
    description = Column(String(100))
    species = relationship("Species", back_populates="mgmt_type")

class ManagementStrategy(Base):
    __tablename__ = 'mgmt_strategy'
    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(10), unique=True)
    description = Column(String(100))
    species = relationship("Species", back_populates="mgmt_strategy")

class Climate(Base):
    __tablename__ = 'climate'
    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(10), unique=True)
    description = Column(String(100))
    species = relationship("Species", back_populates="climate")

class ConBroad(Base):
    __tablename__ = 'con_broad'
    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(10), unique=True)
    description = Column(String(100))
    species = relationship("Species", back_populates="con_broad")

class SiteIndex(Base):
    __tablename__ = 'site_index'
    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(10), unique=True)
    description = Column(String(100))
    species = relationship("Species", back_populates="site_index")

class GrowthPeriod(Base):
    __tablename__ = 'growth_period'
    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(10), unique=True)
    description = Column(String(100))

class Disturbance(Base):
    __tablename__ = 'disturbance'
    id = Column(Integer, primary_key=True, autoincrement=True)
    code = Column(String(10), unique=True)
    description = Column(String(100))
    simulations = relationship("Simulation", back_populates="disturbance")



class Species(Base):

    __tablename__ = 'species'

    id = Column(Integer, primary_key=True, autoincrement=True)
    status_id = Column(Integer, ForeignKey('status.id'))
    forest_type_id = Column(Integer, ForeignKey('forest_type.id'))
    region_id = Column(Integer, ForeignKey('region.id'))
    mgmt_type_id = Column(Integer, ForeignKey('mgmt_type.id'))
    mgmt_strategy_id = Column(Integer, ForeignKey('mgmt_strategy.id'))
    climate_id = Column(Integer, ForeignKey('climate.id'))
    con_broad_id = Column(Integer, ForeignKey('con_broad.id'))
    site_index_id = Column(Integer, ForeignKey('site_index.id'))

    status = relationship("Status", back_populates="species")
    forest_type = relationship("ForestType", back_populates="species")
    region = relationship("Region", back_populates="species")
    mgmt_type = relationship("ManagementType", back_populates="species")
    mgmt_strategy = relationship("ManagementStrategy", back_populates="species")
    climate = relationship("Climate", back_populates="species")
    con_broad = relationship("ConBroad", back_populates="species")
    site_index = relationship("SiteIndex", back_populates="species")
    simulations = relationship("Simulation", back_populates="species")


class Simulation(Base):

    __tablename__ = 'simulation'

    id = Column(Integer, primary_key=True, autoincrement=True)
    species_id = Column(Integer, ForeignKey('species.id'))
    disturbance_id = Column(Integer, ForeignKey('disturbance.id'))
    start = Column(Integer)
    index = Column(Integer)

    species = relationship("Species", back_populates="simulations")
    disturbance = relationship("Disturbance", back_populates="simulations")



class Database:

    def __init__(self, path:str):

        # create engine
        self.engine = create_engine(path)

        # enable foreign key constraints for SQLite
        if path.startswith('sqlite'):
            self._enable_sqlite_foreign_keys()

        # create session
        Session = sessionmaker(bind=self.engine)
        self.session = Session()


    def _enable_sqlite_foreign_keys(self):

        from sqlalchemy import event
        @event.listens_for(self.engine, "connect")
        def enable_sqlite_foreign_keys(dbapi_connection, connection_record):
            cursor = dbapi_connection.cursor()
            cursor.execute("PRAGMA foreign_keys=ON")
            cursor.close()


    def init(self,
             classifiers:dict[str,dict[str,str]],
             disturbances:dict[str,str],
            ):

        # create db
        Base.metadata.create_all(self.engine)

        # populate classifiers table
        models = {'status': Status,
                  'forest_type': ForestType,
                  'region': Region,
                  'mgmt_type': ManagementType,
                  'mgmt_strategy': ManagementStrategy,
                  'climate': Climate,
                  'con_broad': ConBroad,
                  'site_index': SiteIndex,
                  'growth_period': GrowthPeriod,
                 }
        records = []
        self.classif_ids = {}
        for key, values in classifiers.items():
            if key not in models:
                continue
            record = models[key]
            self.classif_ids[key] = {}
            for idx, (code, description) in enumerate(values.items()):
                records.append(record(code=code, description=description))
                self.classif_ids[key][code] = idx+1

        # populate disturbances table
        self.disturbance_ids = {}
        for idx, (code, description) in enumerate(disturbances.items()):
            records.append(Disturbance(code=code, description=description))
            self.disturbance_ids[code] = idx+1

        # commit
        self.session.add_all(records) 
        self.session.commit()


    def classifier_ids(self, table:str)->dict[str,str]:
        return {key:str(value) for key, value in self.classif_ids[table].items()}


    def add_commit_species(self, species:dict)->int:
        record = Species(
            status_id        = self.classif_ids['status'][species['status']],
            forest_type_id   = self.classif_ids['forest_type'][species['forest_type']],
            region_id        = self.classif_ids['region'][species['region']],
            mgmt_type_id     = self.classif_ids['mgmt_type'][species['mgmt_type']],
            mgmt_strategy_id = self.classif_ids['mgmt_strategy'][species['mgmt_strategy']],
            climate_id       = self.classif_ids['climate'][str(species['climate'])],
            con_broad_id     = self.classif_ids['con_broad'][species['con_broad']],
            site_index_id    = self.classif_ids['site_index'][str(species['site_index'])],
        )
        self.session.add(record)
        self.session.commit()
        return record.id


    def add_simulation(self, species_id:int, disturbance_code:int, start:int, index:int):
        record = Simulation(
            species_id     = species_id,
            disturbance_id = self.disturbance_ids[disturbance_code],
            start          = start,
            index          = index,
        )
        self.session.add(record)
        self.session.commit()


    def commit(self):
        self.session.commit()


    def get_simulations(self,
                        status=None,
                        forest_type=None,
                        region=None,
                        mgmt_type=None,
                        mgmt_strategy=None,
                        climate=None,
                        con_broad=None,
                        site_index=None,
                        disturbance=None)->pandas.DataFrame:

        query = self.session.query(Simulation).join(Simulation.species)
        if disturbance:
            query = query.join(Simulation.disturbance).filter(Disturbance.code == disturbance)
        if status:
            query = query.join(Species.status).filter(Status.code == status)
        if forest_type:
            query = query.join(Species.forest_type).filter(ForestType.code == forest_type)
        if region:
            query = query.join(Species.region).filter(Region.code == region)
        if mgmt_type:
            query = query.join(Species.mgmt_type).filter(ManagementType.code == mgmt_type)
        if mgmt_strategy:
            query = query.join(Species.mgmt_strategy).filter(ManagementStrategy.code == mgmt_strategy)
        if climate:
            query = query.join(Species.climate).filter(Climate.code == climate)
        if con_broad:
            query = query.join(Species.con_broad).filter(ConBroad.code == con_broad)
        if site_index:
            query = query.join(Species.site_index).filter(SiteIndex.code == site_index)
        sims = query.all()

        sims = [(sim.species.status.code,
                 sim.species.forest_type.code,
                 sim.species.region.code,
                 sim.species.mgmt_type.code,
                 sim.species.mgmt_strategy.code,
                 sim.species.climate.code,
                 sim.species.con_broad.code,
                 sim.species.site_index.code,
                 sim.disturbance.code,
                 sim.start,
                 sim.index) for sim in sims]

        df = pandas.DataFrame(sims, columns=['status', 'forest_type', 'region', 'mgmt_type', 'mgmt_strategy', 'climate', 'con_broad', 'site_index', 'disturbance', 'start', 'index'])
        return df


    def get_classifiers(self, table_name:str)->pandas.DataFrame:

        match table_name:
            case 'status':
                records = self.session.query(Status).all()
            case 'forest_type':
                records = self.session.query(ForestType).all()
            case 'region':
                records = self.session.query(Region).all()
            case 'mgmt_type':
                records = self.session.query(ManagementType).all()
            case 'mgmt_strategy':
                records = self.session.query(ManagementStrategy).all()
            case 'climate':
                records = self.session.query(Climate).all()
            case 'con_broad':
                records = self.session.query(ConBroad).all()
            case 'site_index':
                records = self.session.query(SiteIndex).all()
            case 'growth_period':
                records = self.session.query(GrowthPeriod).all()
            case 'disturbance':
                records = self.session.query(Disturbance).all()
            case _:
                raise ValueError(f"Unknown table: {table_name}")

        codes = [(record.code, record.description) for record in records]
        return pandas.DataFrame(codes, columns=['code', 'description'])


    def close(self):
        self.session.close()


import pandas
from libcbm.model.cbm.cbm_variables import CBMVariables

def cbm_vars_to_df(cbmvariables: CBMVariables, df_name:str) -> pandas.DataFrame:
    """Extract a data frame from cbm_vars and rename columns if it's the inventory"""
    df = getattr(cbmvariables, df_name).to_pandas()
    # The age and land_class columns appears twice, rename those in the inventory
    if df_name == "inventory":
        df.rename(columns = {'age':        'inv_start_age',
                             'land_class': 'inv_start_land_class'}, inplace=True)
    return df



from functools import cached_property
import numpy as np
import re
import warnings
import pandas
from eu_cbm_hat.post_processor.hwp_common_input import hwp_common_input
from eu_cbm_hat.info.silviculture import keep_clfrs_without_question_marks


class HWP:
    """Compute the Harvested Wood Products Sink

    Example usage:

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['reference'].runners['LU'][-1]
        >>> runner.post_processor.irw_frac
        >>> runner.post_processor.hwp.fluxes_to_products
        >>> runner.post_processor.hwp.fluxes_to_irw
        >>> runner.post_processor.hwp.fluxes_by_age_to_dbh
        >>> runner.post_processor.hwp.fluxes_by_grade_dbh
        >>> runner.post_processor.hwp.fluxes_by_grade
        >>> runner.post_processor.hwp.build_hwp_stock

    Compute results for the default scenario and for another hwp scenario

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['reference'].runners['LU'][-1]
        >>> hwp = runner.post_processor.hwp
        >>> print("Fractions before modification, in the default scenario")
        >>> print(hwp.fraction_semifinished_n_years_mean)
        >>> print(hwp.fraction_semifinished)
        >>> print(hwp.prod_from_dom_harv_sim)
        >>> print(hwp.stock_sink_results)

    Change the fraction semi finished products

        >>> runner.post_processor.hwp.hwp_frac_scenario = "more_sw"
        >>> print("Fractions in the scenario")
        >>> print(hwp.fraction_semifinished)
        >>> # Display the effect on the production from domestic harvest
        >>> print(hwp.prod_from_dom_harv_sim)
        >>> # Display the effect on the final results
        >>> print(hwp.stock_sink_results)

    Set export import factors to one. In other words, assume that all secondary
    products production is made from domestic industrial roundwood harvest.
        
        >>> hwp = runner.post_processor.hwp
        >>> print("no_export_no_import:", hwp.no_export_no_import)
        >>> hwp.no_export_no_import = False
        >>> print(hwp.prod_from_dom_harv_stat)
        >>> print(hwp.stock_sink_results)
        >>> # Change the export import factors to one
        >>> hwp.no_export_no_import = True
        >>> print(hwp.prod_from_dom_harv_stat)
        >>> print(hwp.stock_sink_results)

    TODO:

        - Illustrate change of number of years used to compute domestic factors
          self.n_years_dom_frac = 10

    """

    def __init__(self, parent):
        self.parent = parent
        self.runner = parent.runner
        self.combo_name = self.runner.combo.short_name
        self.classif_list = self.parent.classifiers_list
        # IRW fractions
        self.irw_frac = self.parent.irw_frac
        # Use pool fluxes to get area and age class as well
        self.pools_fluxes = self.runner.output.pool_flux
        # Number of common years to be used to compute the
        # Fraction domestic semi finished products
        self.n_years_dom_frac = 3
        self.hwp_frac_scenario = "default"
        # Add recycling information or not
        self.add_recycling = True
        # Set export import factors to one
        self.no_export_no_import = False

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.runner.short_name)

    @property
    def prod_from_dom_harv_stat(self) -> pandas.DataFrame:
        """Production from domestica hravest statistiscs from hwp_common_input"""
        hwp_common_input.no_export_no_import = self.no_export_no_import
        return hwp_common_input.prod_from_dom_harv_stat

    @cached_property
    def fluxes_to_products(self) -> pandas.DataFrame:
        """Fluxes to products retain from the cbm output the all transfers to
        products pool Remove lines where there are no fluxes to products. Keep
        only lines with positive flues.

            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['reference'].runners['LU'][-1]
            >>> runner.post_processor.hwp.fluxes_to_products
        """
        index_cols = ["year", "area", "disturbance_type", "age_class"]
        fluxes_cols = [
            "softwood_merch_to_product",
            "softwood_other_to_product",
            "softwood_stem_snag_to_product",
            "softwood_branch_snag_to_product",
            "hardwood_merch_to_product",
            "hardwood_other_to_product",
            "hardwood_stem_snag_to_product",
            "hardwood_branch_snag_to_product",
        ]
        cols_of_interest = self.classif_list + index_cols + fluxes_cols
        df = self.pools_fluxes[cols_of_interest]
        # Keep only lines where there are fluxes to products.
        selector = df[fluxes_cols].sum(axis=1) > 0
        df = df.loc[selector].reset_index(drop=True)
        # Merge with IRW fractions
        clfrs_noq = keep_clfrs_without_question_marks(self.irw_frac, self.classif_list)
        df = df.merge(
            self.irw_frac,
            how="left",
            on=clfrs_noq + ["disturbance_type", "year"],
            suffixes=("", "_irw_frac_1"),
        )
        return df

    @cached_property
    def fluxes_to_irw(self) -> pandas.DataFrame:
        """Fluxes to Industrial Roundwood Aggregated by index Extract the IRW
        only, e.g. separate the df as IRW ub exclude bark, because cbm output
        includes the barkand add it to FW. E.g., "df ['softwood_merch']" is the
        IRW fraction.
        """
        df = self.fluxes_to_products
        # Add bark fraction
        df = df.merge(self.parent.wood_density_bark_frac, on="forest_type", how="left")

        # fmt: off
        # Keep only the IRW fraction by multiplying with the fractions coming from self.irw_frac
        df["tc_soft_irw_merch"] = (df["softwood_merch_to_product"] * df["softwood_merch_irw_frac"] * (1 - df["bark_frac"]))
        df["tc_soft_irw_other"] = (df["softwood_other_to_product"] * df["softwood_other_irw_frac"] * (1 - df["bark_frac"]))
        df["tc_soft_irw_stem_snag"] = (df["softwood_stem_snag_to_product"] * df["softwood_stem_snag_irw_frac"] * (1 - df["bark_frac"]))
        df["tc_soft_irw_branch_snag"] = (df["softwood_branch_snag_to_product"] * df["softwood_branch_snag_irw_frac"] * (1 - df["bark_frac"]))

        df["tc_hard_irw_merch"] = (df["hardwood_merch_to_product"] * df["hardwood_merch_irw_frac"] * (1 - df["bark_frac"]))
        df["tc_hard_irw_other"] = (df["hardwood_other_to_product"] * df["hardwood_other_irw_frac"] * (1 - df["bark_frac"]))
        df["tc_hard_irw_stem_snag"] = (df["hardwood_stem_snag_to_product"] * df["hardwood_stem_snag_irw_frac"] * (1 - df["bark_frac"]))
        df["tc_hard_irw_branch_snag"] = (df["hardwood_branch_snag_to_product"] * df["hardwood_branch_snag_irw_frac"] * (1 - df["bark_frac"]))
        # fmt: on

        # Aggregate
        index = [
            "year",
            "forest_type",
            "mgmt_type",
            "mgmt_strategy",
            "con_broad",
            "age_class",
        ]
        tc_cols = df.columns[df.columns.str.contains("tc_")]
        # Aggregate over the index
        df_agg = df.groupby(index)[tc_cols].agg("sum")
        # Sum fluxes columns together into one tc_irw column
        df_agg = df_agg[tc_cols].sum(axis=1).reset_index()
        df_agg.rename(columns={0: "tc_irw"}, inplace=True)
        return df_agg

    @cached_property
    def fluxes_by_age_to_dbh(self) -> pandas.DataFrame:
        """Allocate fluxes by age to a dbh_alloc distrubution

        dbh_allo is the allocation data frame. We merge it with the fluxes to irw.
        """
        # Select data for one country only
        dbh_alloc = hwp_common_input.irw_allocation_by_dbh
        selector = dbh_alloc["country"] == self.runner.country.iso2_code
        dbh_alloc = dbh_alloc.loc[selector]
        # Merge with fluxes
        index = ["mgmt_type", "mgmt_strategy", "age_class", "forest_type"]
        df = self.fluxes_to_irw.merge(dbh_alloc, on=index, how="left")
        # Multiply old tc_irw with the fraction
        df["tc_irw"] = df["tc_irw"] * df["fraction_theoretical_volume"]
        # Reaggregate and check that values didn't change
        index = [
            "year",
            "forest_type",
            "mgmt_type",
            "mgmt_strategy",
            "con_broad",
            "age_class",
        ]
        df_agg = df.groupby(index)["tc_irw"].agg("sum")
        df_comp = self.fluxes_to_irw.merge(
            df_agg, on=index, how="left", suffixes=("_before", "_after")
        )
        selector = (df_comp["year"] > 2020) & (
            ~np.isclose(df_comp["tc_irw_before"], df_comp["tc_irw_after"])
        )
        if any(selector):
            msg = f"\n For post-2020, the following tc_irw values don't match between "
            msg += "input before and after dbh allocation\n"
            msg += f"i.e., allocation on the specified age_class missing because incomplete input in irw_allocation_by_dbh.csv\n"
            msg += f" {df_comp.loc[selector]}"
            warnings.warn(msg)
        return df

    @cached_property
    def fluxes_by_grade_dbh(self) -> pandas.DataFrame:
        """Allocate fluxes by age to a dbh_alloc distribution

        Returns a proportion by species, grade and db_class and also tc_irw.
        """
        # Select the country for nb grading
        nb_grading = hwp_common_input.nb_grading
        selector = nb_grading["country"] == self.runner.country.iso2_code
        nb_grading = nb_grading.loc[selector]
        # Aggregate previous data frame by genus
        df = self.fluxes_by_age_to_dbh
        index = [
            "country",
            "year",
            "genus",
            "mgmt_type",
            "mgmt_strategy",
            "con_broad",
            "dbh_class",
        ]
        df_agg = df.groupby(index)["tc_irw"].agg("sum").reset_index()
        # Merge with grading information
        index = ["country", "genus", "mgmt_type", "mgmt_strategy", "dbh_class"]
        df2 = df_agg.merge(nb_grading, on=index, how="left")
        # Compute the allocation
        df2["tc_irw"] = df2["tc_irw"] * df2["proportion"]
        # Check that values didn't change before and after the allocation
        index = [
            "country",
            "year",
            "genus",
            "mgmt_type",
            "mgmt_strategy",
            "con_broad",
            "dbh_class",
        ]
        df_agg2 = df2.groupby(index).agg(
            tc_irw=("tc_irw", "sum"), proportion=("proportion", "sum")
        )
        df_comp = df_agg.merge(
            df_agg2, on=index, how="left", suffixes=("_before", "_after")
        )
        selector = (df_comp["year"] > 2020) & (
            ~np.isclose(df_comp["tc_irw_before"], df_comp["tc_irw_after"])
        )
        if any(selector):
            msg = f"For post-2020, the following tc_irw values don't match between "
            msg += "input before and after NB grading allocation\n"
            msg = f"For post-2020, fractions for allocation on saw and pulplogs are missing for specified dbh_class"
            msg += f"{df_comp.loc[selector]}"
            warnings.warn(msg)
        return df2

    @cached_property
    def fluxes_by_grade(self) -> pandas.DataFrame:
        """Aggregate previous data frame and reshape wide by feedstock"""
        index = ["year", "grade", "con_broad"]
        df_long = (
            self.fluxes_by_grade_dbh.groupby(index)["tc_irw"].agg("sum").reset_index()
        )
        # Glue grade and con_broad columns together
        df_long["grade2"] = df_long["grade"] + "_" + df_long["con_broad"]
        df = df_long.pivot(
            columns="grade2", index=["year"], values="tc_irw"
        ).reset_index()
        # Sum sawlogs broad and con together
        df["sawlogs"] = df["sawlogs_con"] + df["sawlogs_broad"]
        # Sum back pulpwood con and broad together
        df["pulpwood"] = df["pulpwood_con"] + df["pulpwood_broad"]
        # Remove pulpwood broad and con
        df.drop(columns=["pulpwood_con", "pulpwood_broad"], inplace=True)
        return df

    @property  # Don't cache, in case we change the number of years
    def fraction_semifinished_n_years_mean(self) -> pandas.DataFrame:
        """Compute the fraction of semi finished products as the average of the
        n years defined as in the n_years_dom_frac property

        Also compute the average of the absolute amounts of recycled wood
        entering in wood panels and the amount of recycled paper entering in
        paper production. So that we don't overestimate the contribution of
        fresh forest fibre from domestic production.

        Merge country statistics on domestic harvest and
        CBM output for n common years.

        Check if available raw material is sufficient to produce the amount of
        semi finished products reported by countries.

        param n: Number of years to keep from domestic harvest

        Example use using 3 or 4 years:

            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['reference'].runners['LU'][-1]
            >>> runner.post_processor.hwp.n_years_dom_frac = 3
            >>> print(runner.post_processor.hwp.fraction_semifinished_n_years_mean)
            >>> runner.post_processor.hwp.n_years_dom_frac = 4
            >>> print(runner.post_processor.hwp.fraction_semifinished_n_years_mean)

        """
        # Country statistics on domestic harvest
        dstat = self.prod_from_dom_harv_stat
        # CBM output
        df_out = self.fluxes_by_grade
        index = ["area", "year"]
        cols = [
            "sw_broad_dom_tc",
            "sw_con_dom_tc",
            "wp_dom_tc",
            "pp_dom_tc",
            "recycled_paper_prod",
            "recycled_wood_prod",
        ]
        # Keep data for the last n years and for the selected country
        selector = dstat["year"] > dstat["year"].max() - self.n_years_dom_frac
        selector &= dstat["area"] == self.runner.country.country_name
        dstat = dstat.loc[selector, index + cols]
        # Merge country statistics with CBM output
        df = df_out.merge(dstat, on="year", how="right")
        # calculate the fractions for n years available in case, simulation is
        # based on absolute amounts required in future, then df["sw_dom_tc"],
        # df["pp_dom_tc"], df["wp_dom_tc"] have to be generated from that input
        # data just before the following arithmetic's
        df["sw_broad_fraction"] = df["sw_broad_dom_tc"] / df["sawlogs_broad"]
        df["sw_con_fraction"] = df["sw_con_dom_tc"] / df["sawlogs_con"]
        df["pp_fraction"] = df["pp_dom_tc"] / df["pulpwood"]
        df["sw_dom_tc"] = df["sw_broad_dom_tc"] + df["sw_con_dom_tc"]
        df["wp_fraction"] = df["wp_dom_tc"] / (
            (df["sawlogs"] - df["sw_dom_tc"]) + (df["pulpwood"] - df["pp_dom_tc"])
        )

        # Check if available raw material is sufficient to produce the amount
        # of semi finished products reported by countries.
        # Roundwood can never be converted totally to sawnwood. Fraction always have to
        # be below this value.
        sw_selector = df["sw_broad_fraction"] > 0.7
        if any(sw_selector):
            msg = "Check broad sawnwood production from "
            msg += "sawlogs production for the following years:\n"
            msg += f"{df.loc[sw_selector]}"
            msg += "\nThis temporary warning related to the sw_broad_fraction "
            msg += "should be an error instead."
            warnings.warn(msg)

        # Roundwood can never be converted totally to sawnwood. Fraction always have to
        # be below this value.
        sw_selector = df["sw_con_fraction"] > 0.6
        if any(sw_selector):
            msg = "Check con sawnwood production from"
            msg += "sawlogs production for the following years:\n"
            msg += f"{df.loc[sw_selector]}"
            raise ValueError(msg)

        pp_selector = df["pp_fraction"] > 1
        if any(pp_selector):
            msg = "Check paper production from"
            msg += "pulpwood/pulplogs production for the following years:\n"
            msg += f"{df.loc[pp_selector]}"
            raise ValueError(msg)

        wp_selector = df["wp_fraction"] > 1
        if any(wp_selector):
            msg = "Check wood panels production from"
            msg += ("pulpwood and sawnwood production for the following years:\n")
            msg += f"{df.loc[wp_selector]}"
            raise ValueError(msg)

        # Compute the average of the selected columns
        selected_cols = [
            'sw_broad_fraction',
            'sw_con_fraction',
            "pp_fraction",
            "wp_fraction",
            "recycled_paper_prod",
            "recycled_wood_prod",
        ]
        mean_frac = df[selected_cols].mean()
        return mean_frac

    @property  # Don't cache, in case we change the number of years
    def fraction_semifinished_default(self) -> pandas.DataFrame:
        """Fraction of semi finished products in the default case"""
        mean_frac = self.fraction_semifinished_n_years_mean
        max_year = self.runner.country.base_year + self.runner.num_timesteps
        df = pandas.DataFrame({"year": range(1900, max_year + 1)})
        cols = [
            'sw_broad_fraction',
            'sw_con_fraction',
            "pp_fraction",
            "wp_fraction",
            "recycled_paper_prod",
            "recycled_wood_prod",
        ]
        for col in cols:
            df[col] = mean_frac[col]
        # Default recycling values to one
        df["recycled_wood_factor"] = 1
        df["recycled_paper_factor"] = 1
        # Return
        return df

    @property  # Don't cache, in case we change the number of years
    def fraction_semifinished_scenario(self) -> pandas.DataFrame:
        """Fraction of semi finished products scenario case"""
        mean_frac = self.fraction_semifinished_n_years_mean
        df = hwp_common_input.hwp_fraction_semifinished_scenario.copy()
        selector = df["country"] == self.runner.country.country_name
        selector &= df["hwp_frac_scenario"] == self.hwp_frac_scenario
        df.drop(columns=["country", "hwp_frac_scenario"], inplace=True)
        # Add recycled values from the mean table
        df["recycled_paper_prod"] = mean_frac["recycled_paper_prod"]
        df["recycled_wood_prod"] = mean_frac["recycled_wood_prod"]
        return df

    @property  # Don't cache, in case we change the number of years
    def fraction_semifinished(self) -> pandas.DataFrame:
        """Fraction of semi finished products

        Either default values or values from a scenario input files:

        - The default values are from the fraction_semifinished_n_years_mean
          method
        - The input files values are from
          hwp_common_input.hwp_fraction_semifinished_scenario
        """
        if self.hwp_frac_scenario == "default":
            # Default scenario
            return self.fraction_semifinished_default
        # Other scenarios
        return self.fraction_semifinished_scenario

    # Don"t cache, in case we change the number of years or the self.add_recycling
    # property
    @property
    def prod_from_dom_harv_sim(self) -> pandas.DataFrame:
        """Compute the production of sanwood, panels and paper from domestic
        harvest as an output of the CBM simulated amounts of sawlogs and
        pulpwood.

        TODO: Enable scenarios of production from domestic harvest from an
        external model. See issue 104.
        https://gitlab.com/bioeconomy/eu_cbm/eu_cbm_hat/-/issues/104#top

        Illustrate a change of the sw, wp, pp fractions:

            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['reference'].runners['LU'][-1]
            >>> hwp = runner.post_processor.hwp
            >>> print("Fractions before modification, in the default scenario")
            >>> print(hwp.fraction_semifinished_n_years_mean)
            >>> print(hwp.fraction_semifinished)
            >>> print(hwp.prod_from_dom_harv_sim)
            >>> print(hwp.stock_sink_results)

            >>> # Change the fraction semi finished
            >>> runner.post_processor.hwp.hwp_frac_scenario = "more_sw"
            >>> print("Fractions in the scenario")
            >>> print(hwp.fraction_semifinished)
            >>> # Display the effect on the production from domestic harvest
            >>> print(hwp.prod_from_dom_harv_sim)
            >>> # Display the effect on the final results
            >>> print(hwp.stock_sink_results)

            >>> # Add recycling (default)
            >>> df_with_recycling = hwp.prod_from_dom_harv_sim
            >>> # Don't add recycling
            >>> hwp.add_recycling = False
            >>> df_without_recycling = hwp.prod_from_dom_harv_sim

        """
        df = self.fluxes_by_grade.copy()
        # Add the fractions to the CBM output data
        df = df.merge(self.fraction_semifinished, on="year")
        # Compute production from domestic harvest for the future
        df["sw_broad_dom_tc"] = df["sawlogs_broad"] * df["sw_broad_fraction"]
        df["sw_con_dom_tc"] = df["sawlogs_con"] * df["sw_con_fraction"]
        df["wp_dom_tc"] = df["sawlogs"] * df["wp_fraction"]
        df["pp_dom_tc"] = df["pulpwood"] * df["pp_fraction"]
        if self.add_recycling:
            msg = "Add recycling amounts because "
            msg += f"add_recycling = {self.add_recycling}"
            print(msg)
            df["wp_dom_tc"] += (
                df["recycled_wood_prod"]
                * hwp_common_input.c_wp
                * df["recycled_wood_factor"]
            )
            df["pp_dom_tc"] += (
                df["recycled_paper_prod"]
                * hwp_common_input.c_pp
                * df["recycled_paper_factor"]
            )

##########
        #### NEED to check if these are smaller than the corresponding 'loss' in the year
            #df["recycled_wood_prod"] 
            #* hwp_common_input.c_wp
            #* df["recycled_wood_factor"] +
            #        df["recycled_paper_prod"]
            #        * hwp_common_input.c_pp
            #        * df["recycled_paper_factor"] < df['hwp_loss']

##########
        
        else:
            msg = "No recycling amounts because "
            msg += f"add_recycling = {self.add_recycling}"
            print(msg)
        return df

    @property  # Don't cache, in case we change the number of years
    def concat_1900_to_last_sim_year(self) -> pandas.DataFrame:
        """This applies with IPCC method. Concatenate backcasted data from
        1900, reported historical country data and CBM output until the end of
        the simulation.

        In the CBM simulated output, keep only years that are not already in
        country statistics.

        Plot production from domestic harvest for both historically reported
        data and simulation output:

            >>> import matplotlib.pyplot as plt
            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['reference'].runners['LU'][-1]
            >>> df = runner.post_processor.hwp.concat_1900_to_last_sim_year
            >>> df.set_index("year").plot()
            >>> plt.show()

        """
        # Input data frames
        dstat = self.prod_from_dom_harv_stat.copy()
        df_out = self.prod_from_dom_harv_sim.copy()
        cols = ['sw_broad_dom_tc', 'sw_con_dom_tc', "wp_dom_tc", "pp_dom_tc"]
        # Keep data for the selected country
        selector = dstat["area"] == self.runner.country.country_name
        dstat = dstat.loc[selector, ["year"] + cols]
        df_out = df_out[["year"] + cols].copy()
        # Keep only years not in country statistics
        selector = df_out["year"] > dstat["year"].max()
        df_out = df_out.loc[selector]
        # Concatenate
        df = pandas.concat([dstat, df_out])
        return df.reset_index(drop=True)

    @property  # Don't cache, in case we change the number of years
    def prepare_decay_and_inflow(self):
        """Prepare decay parameters and compute inflow"""
        df = self.concat_1900_to_last_sim_year.copy()
        # Assign decay parameters inside df
        decay_params = hwp_common_input.decay_params
        for col in decay_params.columns:
            df[col] = decay_params[col].values[0]
        # Compute the corrected inflow according to decay elements in the square bracket
        # Estimate the annual inflows
        df = df.assign(
            sw_broad_inflow=(df.sw_broad_dom_tc * df.k1_sw),
            sw_con_inflow=(df.sw_con_dom_tc * df.k1_sw),
            wp_inflow=(df.wp_dom_tc * df.k1_wp),
            pp_inflow=(df.pp_dom_tc * df.k1_pp),
        )
        return df

    @property  # Don't cache, in case we change the number of years
    def build_hwp_stock_since_1900(self):
        # def build_hwp_stock(self):
        """IPCC method: Build HWP stock values for 1900 to end of simulated period

        Plot stock evolution

            >>> import matplotlib.pyplot as plt
            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['reference'].runners['LU'][-1]
            >>> df = runner.post_processor.hwp.build_hwp_stock_since_1900
            >>> cols = ['sw_stock', 'wp_stock', 'pp_stock']
            >>> df.set_index("year")[cols].plot()
            >>> plt.show()

        """
        df = self.prepare_decay_and_inflow.copy()
        cols = ["sw_broad_inflow","sw_con_inflow", "wp_inflow", "pp_inflow"]
        cols += ["e_sw", "e_wp", "e_pp", "k_sw", "k_wp", "k_pp"]
        df = df[["year"] + cols]
        # Initiate stock values
        df["sw_broad_stock"] = 0.0  # Keep the dot to create a floating point number
        df["sw_con_stock"] = 0.0
        df["wp_stock"] = 0.0
        df["pp_stock"] = 0.0

        # Compute the stock for each semi finite product for all subsequent years
        df = df.set_index("year")
        for t in range(df.index.min() + 1, df.index.max()):
            df.loc[t, "sw_broad_stock"] = (
                df.loc[t - 1, "sw_broad_stock"] * df.loc[t, "e_sw"] + df.loc[t, "sw_broad_inflow"]
            )
            df.loc[t, "sw_con_stock"] = (
                df.loc[t - 1, "sw_con_stock"] * df.loc[t, "e_sw"] + df.loc[t, "sw_con_inflow"]
            )
            df.loc[t, "wp_stock"] = (
                df.loc[t - 1, "wp_stock"] * df.loc[t, "e_wp"] + df.loc[t, "wp_inflow"]
            )
            df.loc[t, "pp_stock"] = (
                df.loc[t - 1, "pp_stock"] * df.loc[t, "e_pp"] + df.loc[t, "pp_inflow"]
            )
        df.reset_index(inplace=True)
        df.fillna(0, inplace=True)

        # Add annual loses by decay of historical stock, as the limit for recycling
        df["sw_con_loss"] = df["sw_con_stock"] * df["k_sw"]
        df["sw_broad_loss"] = df["sw_broad_stock"] * df["k_sw"]
        df["wp_loss"] = df["wp_stock"] * df["k_wp"]
        df["pp_loss"] = df["pp_stock"] * df["k_pp"]
        df["hwp_loss"] = df["sw_con_loss"] + df["sw_broad_loss"] + df["wp_loss"] + df["pp_loss"]

        # Compute the total stock
        df["hwp_tot_stock_tc"] = df["sw_broad_stock"] + df["sw_broad_stock"] + df["wp_stock"] + df["pp_stock"]

        # Do the difference between consecutive years
        df["hwp_tot_diff_tc"] = df["hwp_tot_stock_tc"].diff(periods=1)
        # Stock diff shifted by one year
        # df['hwp_tot_diff_tc_m1'] = df['hwp_tot_stock_tc'].diff(periods=1).shift(-1)
        df["hwp_tot_sink_tco2"] = df["hwp_tot_diff_tc"] * (-44 / 12)
        return df

    @property  # Don't cache, in case we change the number of years
    def build_hwp_stock_since_1990(self):
        """complementary KP method: build HWP stock values for 1990 to the end
        of simulated period

        Plot stock evolution

            >>> import matplotlib.pyplot as plt
            >>> from eu_cbm_hat.core.continent import continent
            >>> runner = continent.combos['reference'].runners['LU'][-1]
            >>> df = runner.post_processor.hwp.build_hwp_stock_1990
            >>> cols = ['sw_stock', 'wp_stock', 'pp_stock']
            >>> df.set_index("year")[cols].plot()
            >>> plt.show()

        """

        df = self.prepare_decay_and_inflow.copy()
        cols = ["sw_broad_inflow", "sw_con_inflow", "wp_inflow", "pp_inflow"]
        cols += ["k_sw", "k_wp", "k_pp"]
        cols += ["e_sw", "e_wp", "e_pp"]
        df = df[["year"] + cols]
        # Retain only the first five years including 1990
        df_mean = df[df["year"] >= 1990].head(5)
        # Initiate stock values as average of 1990-1995 as the average of the
        # First five years for each inflow type
        df["sw_con_stock"] = (df_mean["sw_con_inflow"].mean()) / (df_mean["k_sw"].mean())
        df["sw_broad_stock"] = (df_mean["sw_broad_inflow"].mean()) / (df_mean["k_sw"].mean())
        df["wp_stock"] = (df_mean["wp_inflow"].mean()) / (df_mean["k_wp"].mean())
        df["pp_stock"] = (df_mean["pp_inflow"].mean()) / (df_mean["k_pp"].mean())

        # Compute the stock for each semi finite product for all subsequent years
        df = df.set_index("year")
        for t in range(df.index.min() + 1, df.index.max()):
            df.loc[t, "sw_broad_stock"] = (
                df.loc[t - 1, "sw_broad_stock"] * df.loc[t, "e_sw"] + df.loc[t, "sw_broad_inflow"]
            )
            df.loc[t, "sw_con_stock"] = (
                df.loc[t - 1, "sw_con_stock"] * df.loc[t, "e_sw"] + df.loc[t, "sw_con_inflow"]
            )
            df.loc[t, "wp_stock"] = (
                df.loc[t - 1, "wp_stock"] * df.loc[t, "e_wp"] + df.loc[t, "wp_inflow"]
            )
            df.loc[t, "pp_stock"] = (
                df.loc[t - 1, "pp_stock"] * df.loc[t, "e_pp"] + df.loc[t, "pp_inflow"]
            )
        df.reset_index(inplace=True)

        # Add annual loses by decay of historical stock, as the limit for recycling
        df["sw_con_loss"] = df["sw_con_stock"] * df["k_sw"]
        df["sw_broad_loss"] = df["sw_broad_stock"] * df["k_sw"]
        df["wp_loss"] = df["wp_stock"] * df["k_wp"]
        df["pp_loss"] = df["pp_stock"] * df["k_pp"]
        df["hwp_loss"] = df["sw_broad_loss"] + df["sw_con_loss"] + df["wp_loss"] + df["pp_loss"]
        # Compute the total stock
        df["hwp_tot_stock_tc"] = df["sw_broad_stock"] + df["sw_con_stock"] + df["wp_stock"] + df["pp_stock"]
        # Do the difference between consecutive years
        df["hwp_tot_diff_tc"] = df["hwp_tot_stock_tc"].diff(periods=1)
        # Stock diff shifted by one year
        # df['hwp_tot_diff_tc_m1'] = df['hwp_tot_stock_tc'].diff(periods=1).shift(-1)
        df["hwp_tot_sink_tco2"] = df["hwp_tot_diff_tc"] * (-44 / 12)
        # Keep only after 1990
        df = df[df["year"] >= 1990]
        return df

    @cached_property
    def fluxes_to_primary_fw(self) -> pandas.DataFrame:
        """Fluxes to primary Fuel Wood."""
        df = self.fluxes_to_products
        # Add bark fraction
        df = df.merge(self.parent.wood_density_bark_frac, on="forest_type", how="left")

        # Switch off black autoformatting to avoid the long lines below to be wrapped.
        # fmt: off
        # Primary Fuel Wood removed directly as fulewood trees
        df['tc_soft_fw_merch'] = df['softwood_merch_to_product'] * (1-df['softwood_merch_irw_frac'])
        df['tc_soft_fw_other'] = df['softwood_other_to_product'] * (1-df['softwood_other_irw_frac'])
        df['tc_soft_fw_stem_snag'] = df['softwood_stem_snag_to_product'] * (1-df['softwood_stem_snag_irw_frac'])
        df['tc_soft_fw_branch_snag'] = df['softwood_branch_snag_to_product'] * (1-df['softwood_branch_snag_irw_frac'])
        df['tc_hard_fw_merch'] = df['hardwood_merch_to_product'] * (1-df['hardwood_merch_irw_frac'])
        df['tc_hard_fw_other'] = df['hardwood_other_to_product'] * (1-df['hardwood_other_irw_frac'])
        df['tc_hard_fw_stem_snag'] = df['hardwood_stem_snag_to_product'] * (1-df['hardwood_stem_snag_irw_frac'])
        df['tc_hard_fw_branch_snag'] = df['hardwood_branch_snag_to_product'] * (1-df['hardwood_branch_snag_irw_frac'])
        # fmt: on

        # Aggregate
        index = ["year"]
        tc_cols = df.columns[df.columns.str.contains("tc_")]
        # Aggregate over the index
        df_agg = df.groupby(index)[tc_cols].agg("sum")
        # Sum fluxes columns together into one tc_irw column
        df_agg = df_agg[tc_cols].sum(axis=1).reset_index()
        df_agg.rename(columns={0: "tc_primary_fw"}, inplace=True)
        return df_agg

    @cached_property
    def fluxes_to_secondary_fw(self) -> pandas.DataFrame:
        """Deduce actual products from overall industrial roundwood

        Part of the eligible HWP amount is converted to actual products. We
        consider that the remaining part is burned as secondary fuel wood.

        Part of the IRW Bark is also used as fuel wood.
        Compute the bark obtain from Industrial Roundwood and add it to the total.

        """
        df = self.fluxes_to_products
        # Add bark fraction
        df = df.merge(self.parent.wood_density_bark_frac, on="forest_type", how="left")

        # fmt: off
        # Compute the bark obtain from Industrial Roundwood
        #IRW's bark to be included in "FW_secondary" pools
        df['tc_soft_fw_irw_merch'] = df['softwood_merch_to_product'] * df ['softwood_merch_irw_frac']*df['bark_frac']
        df['tc_soft_fw_irw_other'] = df['softwood_other_to_product'] * df['softwood_other_irw_frac']*df['bark_frac']
        df['tc_soft_fw_irw_stem_snag'] = df['softwood_stem_snag_to_product'] * df['softwood_stem_snag_irw_frac']*df['bark_frac']
        df['tc_soft_fw_irw_branch_snag'] = df['softwood_branch_snag_to_product'] * df['softwood_branch_snag_irw_frac']*df['bark_frac']
        df['tc_hard_fw_irw_merch'] = df['hardwood_merch_to_product'] * df['hardwood_merch_irw_frac']*df['bark_frac']
        df['tc_hard_fw_irw_other'] = df['hardwood_other_to_product'] * df['hardwood_other_irw_frac']*df['bark_frac']
        df['tc_hard_fw_irw_stem_snag'] = df['hardwood_stem_snag_to_product'] * df['hardwood_stem_snag_irw_frac']*df['bark_frac']
        df['tc_hard_fw_irw_branch_snag'] = df['hardwood_branch_snag_to_product'] * df['hardwood_branch_snag_irw_frac']*df['bark_frac']
        # fmt: on

        cols = [
            "tc_soft_fw_irw_merch",
            "tc_soft_fw_irw_other",
            "tc_soft_fw_irw_stem_snag",
            "tc_soft_fw_irw_branch_snag",
            "tc_hard_fw_irw_merch",
            "tc_hard_fw_irw_other",
            "tc_hard_fw_irw_stem_snag",
            "tc_hard_fw_irw_branch_snag",
        ]

        df = df.groupby(["year"])[cols].sum().reset_index()
        df["fw_irw_bark"] = df[cols].sum(axis=1)
        df1 = self.fluxes_by_grade
        df1["hwp_eligible"] = df1[["pulpwood", "sawlogs"]].sum(axis=1)
        df2 = self.prod_from_dom_harv_sim
        df2["sw_dom_tc"] = df2[["sw_broad_dom_tc", "sw_con_dom_tc"]].sum(axis=1)
        df2["hwp_allocated"] = df2[["sw_dom_tc", "wp_dom_tc", "pp_dom_tc"]].sum(axis=1)
        df = df1[["year", "hwp_eligible"]].merge(
            df2[["year", "hwp_allocated"]].merge(df[["year", "fw_irw_bark"]]), on="year"
        )
        df["tc_secondary_fw"] = (
            df["hwp_eligible"] - df["hwp_allocated"] + df["fw_irw_bark"]
        )

        return df

    @cached_property
    def ghg_emissions_fw(self) -> pandas.DataFrame:
        """Green House Gas Emissions from both primary and secondary fuel wood"""
        df1 = self.fluxes_to_primary_fw
        df2 = self.fluxes_to_secondary_fw
        df = df1.merge(df2, on="year", how="left")
        df["tc_fw"] = df[["tc_primary_fw", "tc_secondary_fw"]].sum(axis=1)

        # convert to joules as EFs are on TJ
        # Net Cal Value by mass: logwood (stacked – air dry: 20% MC) = 14.7 GJ/tonne = 0.0147 Tj/tonne of dry mass (Forestry Commission, 2022).
        ncv = 0.0147  # TJ/tone dry mass
        # EMISSION FACTORS for Biomass category: Wood / Wood Waste.
        # Note that CO2 emissions are already included in living biomass loss (for primary) and in HWP loss (for secondary)
        ef_ch4 = 0.003  # tCH4/TJ
        ef_n2o = 0.0004  # tN2O/TJ
        # GWPs
        gwp_ch4 = 21
        gwp_n2o = 300
        # intermediary calcualtions
        df["ncv_x_ef_ch4_x_gwp"] = ncv * ef_ch4 * gwp_ch4
        df["ncv_x_ef_n2o_x_gwp"] = ef_n2o * gwp_n2o
        # convert to dry mass
        #df["fw_dm"] = df["tc_fw"] / 0.5
        df["fw_primary_dm"] = df["tc_primary_fw"] / 0.5
        df["fw_secondary_dm"] = df["tc_secondary_fw"] / 0.5

        # Estimate the CO2 equivalent emissions for non CO2 gases
        df["fw_primary_ghg_co2_eq"] = (
            df["fw_primary_dm"] * df["ncv_x_ef_ch4_x_gwp"]
            + df["fw_primary_dm"] * df["ncv_x_ef_n2o_x_gwp"]
        )

        df["fw_secondary_ghg_co2_eq"] = (
            df["fw_secondary_dm"] * df["ncv_x_ef_ch4_x_gwp"]
            + df["fw_secondary_dm"] * df["ncv_x_ef_n2o_x_gwp"]
        )   
        df["fw_primary_co2"] = df["tc_primary_fw"] * 44 / 12
        df["fw_secondary_co2"] = df["tc_secondary_fw"] * 44 / 12
        # Estimate the Green House Gas Emissions as CO2 Equivalent
        df["fw_ghg_co2_eq"] =df["fw_primary_ghg_co2_eq"] + df["fw_secondary_ghg_co2_eq"]
        df["fw_co2"] = df["fw_primary_co2"] + df["fw_secondary_co2"]
        
        return df

    @cached_property
    def ghg_emissions_waste(self) -> pandas.DataFrame:
        """Non CO2 emissions from waste dumps Waste amounts
        We don't calculate CO2 emissions because CO2 emissions are already
        accounted under HWP. This is about non-CO2 emissions. Calculation for
        CH4 emissions.

        Implements the method in IPCC 2006-2019. See steps in the Technical report:
            - Extend the time series in the past to prepare to compute
              accumulation
            - Compute cumulated Decomposable Degradable Organic Carbon
            - Get CH4 emissions by appling DOCF and MCF, F
            - Apply GWP Global Warming Potential
        """
        df = hwp_common_input.waste
        # Select data for this country
        selector = df["country_iso2"] == self.runner.country.iso2_code
        df = df.loc[selector]

        # Start from an early year to compute the stock accumulation.
        # Annualize biannual data
        years = np.arange(1960, 2071)
        new_df = pandas.DataFrame({"year": years})

        # Merge the two DataFrames and interpolate
        df = pandas.merge(new_df, df, on="year", how="left")
        df["wood_landfill_tfm"] = df["wood_landfill_tfm"].interpolate()
        df = df.bfill().ffill()
        # Assign decay parameter inside df
        decay_params = hwp_common_input.decay_params
        df["e_ww"] = decay_params["e_sw"].values[0]
        # Initialize cumulated wood waste  data from the annual input
        df["ddoc_mat_doc_stock_tdm"] = 0.0  # Initialize as float instead of Int
        # Initialize annual loss from wood waste stock
        df["ddocm_decompt_tdm"] = 0.0  # Initialize as float instead of Int
        # Add factor for excluding the amount subject to aerobic decomposition
        df["docf_factor"] = 0.5
        # Add factor for mass convertible to CH4
        df["mcf_factor"] = 0.5  #
        df = df.set_index("year")
        # Fill the columns for DDOCM historically cumulated doc, and annual
        # loss ddoc convertible to CH4
        for y in range(df.index.min() + 1, df.index.max() + 1):
            df.loc[y, "ddoc_mat_doc_stock_tdm"] = (
                df.loc[y - 1, "ddoc_mat_doc_stock_tdm"] * df.loc[y, "e_ww"]
                + df.loc[y, "w_annual_wood_landfill_tdm"]
                * df.loc[y, "docf_factor"]
                * df.loc[y, "mcf_factor"]
            )
            df.loc[y, "ddocm_decompt_tdm"] = df.loc[y - 1, "ddoc_mat_doc_stock_tdm"] * (
                1 - df.loc[y, "e_ww"]
            )
        df = df.reset_index()

        # Remove years before a certain period
        df = df[df["year"] > 2020].copy()

        # Get CH4 emissions by appling DOCF and MCF, F
        f_factor = 0.5
        df["ch4_generated_tch4"] = df["ddocm_decompt_tdm"] * f_factor * 16 / 12
        # Apply GWP Global Warming Potential
        df["waste_co2_eq"] = df["ch4_generated_tch4"] * 21
        return df

    @cached_property
    def ctf_unfccc(self) -> pandas.DataFrame:
        """Common Table Format for HWP calibration
        Exogenous reporting data from the CTF country reports to the UNFCCC.
        """
        df = hwp_common_input.ctf_unfccc
        # Select the country
        selector = df["member_state"] == self.runner.country.country_name
        df = df.loc[selector].copy()
        return df

    @property
    def stock_sink_results(self) -> pandas.DataFrame:
        """Comparison table for HWP calibration

        Collect data from different initial year, which correspond to different
        methodologies:

            - Base year 1900 (IPCC 2006/2018/2019)
            - Base year 1990 (IPCC 2013)
            - non-CO2 emissions and total GHG emissions from burning wood for
              energy
            - Load GHG emissions from waste

        """
        # Load GHG emissions reported by country to UNFCCC for HWP pool
        cols = ["member_state", "year", "crf_hwp_tco2"]
        df_ctf = self.ctf_unfccc[cols].copy()
        df_ctf["year"] = df_ctf["year"].astype(int)

        # Load stock and sink
        cols_stock = ["hwp_tot_stock_tc", "hwp_tot_sink_tco2", "hwp_loss"]
        # Initial year 1900 (IPCC 2006/2018/2019)
        df_1900 = self.build_hwp_stock_since_1900[["year"] + cols_stock].copy()
        # Add '_1900' to the columns
        df_1900 = df_1900.rename(columns={col: f"{col}_1900" for col in cols_stock})
        selector = df_1900["year"] >= 1990
        df_1900 = df_1900.loc[selector]

        # Initial year 1990 (IPCC 2013)
        df_1990 = self.build_hwp_stock_since_1990[["year"] + cols_stock]
        # Add '_1990' to the columns
        df_1990 = df_1990.rename(columns={col: f"{col}_1990" for col in cols_stock})

        # Load GHG emissions from burning wood for energy
        cols = ["year","fw_primary_ghg_co2_eq", "fw_secondary_ghg_co2_eq", "fw_primary_co2", "fw_secondary_co2"]
        
        df_fw = self.ghg_emissions_fw[cols]

        # Load GHG emissions from waste
        cols = ["year", "waste_co2_eq"]
        df_waste = self.ghg_emissions_waste[cols]

        # Merge
        df = df_ctf.merge(df_1900, on="year", how="outer")
        df = df.merge(df_1990, on="year", how="outer")
        df = df.merge(df_fw, on="year", how="outer")
        df = df.merge(df_waste, on="year", how="outer")
        # Add a scenario column
        df["hwp_frac_scenario"] = self.hwp_frac_scenario
        # Place the last column first
        cols = df.columns.to_list()
        cols = cols[-1:] + cols[:-1]

        # Fill in the last simulated year with average of the previous two years
        # Define the columns to fill
        cols_to_fill = [
            "hwp_tot_stock_tc_1900",
            "hwp_tot_sink_tco2_1900",
            "hwp_tot_stock_tc_1990",
            "hwp_tot_sink_tco2_1990",
        ]

        # Get the max year
        max_year = df["year"].max()

        # Get the two years before the max year
        prev_years = df[df["year"] < max_year].nlargest(2, "year")["year"].unique()

        # Loop through each column to fill
        for col in cols_to_fill:
            # Calculate the average of the previous two years
            avg_value = df.loc[df["year"].isin(prev_years), col].mean()

            # Fill the missing value with the average
            df.loc[df["year"] == max_year, col] = avg_value

        # Define the columns to keep
        cols_to_keep = [
            "hwp_frac_scenario",
            "member_state",
            "year",
            "crf_hwp_tco2",
            "hwp_tot_stock_tc_1900",
            "hwp_tot_sink_tco2_1900",
            "hwp_tot_stock_tc_1990",
            "hwp_tot_sink_tco2_1990",
            "hwp_loss_1900",
            "hwp_loss_1990",
            "fw_primary_ghg_co2_eq", 
            "fw_secondary_ghg_co2_eq", 
            "fw_primary_co2", 
            "fw_secondary_co2",            
            "waste_co2_eq",
        ]
        # Return the updated DataFrame with all required columns
        df = df[cols_to_keep]

        df["member_state"] = df["member_state"].ffill()

        return df[cols]



"""Modify the growth multiplier for the purpose of taking climate differences into account

Written by Viorel Blujdea and Paul Rougieux.

JRC Biomass Project. Unit D1 Bioeconomy.


Example usage:

    >>> from eu_cbm_hat.core.continent import continent
    >>> runner = continent.combos['reference_cable_pop'].runners['EE'][-1]
    >>> runner.num_timesteps = 2070 - runner.country.inventory_start_year
    >>> # Check availability of the raw growth multiplier table
    >>> runner.clim_adjust.df
    >>> output = runner.run(keep_in_ram=True, verbose=True, interrupt_on_error=True)

    >>> from eu_cbm_hat.core.continent import continent
    >>> runner = continent.combos['reference'].runners['LU'][-1]
    >>> runner.num_timesteps = 2070 - runner.country.inventory_start_year
    >>> output = runner.run(keep_in_ram=True, verbose=True, interrupt_on_error=True)

Inside libcbm's C++ code from the `libcbm_c` repository, we can see where the
growth multiplier is applied inside `src/cbm/cbmbiomassdynamics.cpp`.

There are two multiplications by growth multipliers:

1. one multiplication in GetTotalBiomassIncrement
    - total_increment.SWM = SWM_inc * sw_multiplier;
    - total_increment.HWM = HWM_inc * hw_multiplier;

2. another multiplication in GetGrowthMatrix

```
Biomass inc = GetTotalBiomassIncrement(biomass,
    agBiomassIncrement,
    growthMult.SoftwoodMultiplier * growth_multiplier,
    growthMult.HardwoodMultiplier * growth_multiplier);
```

What is the difference between `growthMult.SoftwoodMultiplier`,
`growth_multiplier` and `sw_multiplier` ?

1. growthMult.SoftwoodMultiplier is a Species and disturbance-specific
adjustment factor retrieved from parameter tables via
`_Parameters.GetGrowthMultipliers(lastDisturbancetype,
growth_multiplier_step)`. It is based on `lastDisturbancetype (e.g., fire,
harvest, insect damage)` and `growth_multiplier_step (likely time since
disturbance)`. It represents how the specific disturbance history affects
softwood growth rates.

2. `growth_multiplier` is an overall growth scaling factor (for climate/site
productivity) passed as a parameter through the cbm_vars.state data frame. It represents
broad environmental conditions affecting all growth such as Climate conditions
(temperature, precipitation), Site quality, CO₂ fertilization effects.

3. `sw_multiplier` is calculated as the product of the above two.
`sw_multiplier = growthMult.SoftwoodMultiplier * growth_multiplier`. It
represents the Total growth adjustment combining both disturbance effects AND
environmental conditions.

Source of the growth modifier inside libcbm_c/src/cbm/cbmdefaultparameters.cpp

            int disturbanceType = t.GetValue(row, "disturbance_type_id");
            int forest_type_id = t.GetValue(row, "forest_type_id");
            int time_step = t.GetValue(row, "time_step");

"""

import warnings
import pandas
from eu_cbm_hat.cbm.cbm_vars_to_df import cbm_vars_to_df
from libcbm.storage import dataframe


class GrowthModifier:
    """Modify the growth multiplier for the purses of taking climate
    differences into account.
    """

    def __init__(self, parent):
        self.parent = parent
        self.runner = parent.runner
        self.combo_name = self.runner.combo.short_name

    def update_state(self, year, cbm_vars):
        """Update the cbm_vars.state with new growth multiplier values

        This method updates the state data frame of cbm vars at the **beginning
        of the time step**, before growth and disturbances are applied.

        The CBM variables used here are based on cbm_vars before the time step.
        This is in contrast to the CBM variables used in `dynamics_func` of
        cbm/dynamic.py. The `stands` data frame in `dynamics_func` is a
        concatenation of the classifiers, parameters, inventory, state, flux
        and pools data frames from `end_vars` which is a simulated result
        of the stand state at the **end of the time step**.
        """
        # Get growth multiplier input data
        clim_adjust_df = self.runner.clim_adjust.df
        # Check if there are growth multipliers values for this year.
        # If not skip and return cbm_vars as is.
        if year not in clim_adjust_df["year"].unique():
            return cbm_vars
        cbm_vars_classif_df = cbm_vars_to_df(cbm_vars, "classifiers")
        cbm_vars_state_df = cbm_vars_to_df(cbm_vars, "state")
        state = pandas.concat([cbm_vars_classif_df, cbm_vars_state_df], axis=1)
        # Keep only the current year
        index = ["climate", "con_broad"]
        selector = clim_adjust_df["year"] == year
        df = clim_adjust_df.loc[selector, index + ["ratio"]].copy()
        # Rename ratio to growth_multiplier
        df.rename(columns={"ratio": "growth_multiplier"}, inplace=True)
        # Convert classifiers IDs from the SIT standard to the user standard
        state = self.parent.conv_clfrs(state)

        # Merge climate adjustment with state keep old columns as "old" and new
        # column as growth_multiplier
        state = state.merge(df, on=index, suffixes=("_old", ""), how="left")

        # Debug message with the new and old values
        msg = "Growth multiplier values from the previous period in CBM state "
        msg += f"{state['growth_multiplier_old'].unique()}"
        self.runner.log.debug(msg)
        msg = "Growth multiplier values adjusted for climate, "
        msg = "computed in runner.clim_adjust.df "
        msg += f"{state['growth_multiplier'].unique()}"
        self.runner.log.debug(msg)

        # Check that the combination of Climate units and con_broad present in
        # CBM are alsoe present in the NPP input file. If not raise a warning
        # and keep growth_multiplier unchanged.
        selector = state["growth_multiplier"].isna()
        if any(selector):
            df_missing = state.loc[
                selector, index + ["growth_multiplier"]
            ].value_counts()
            msg = "Some combinations of climate units and con_broad "
            msg += "are note present int he NPP climate modification "
            msg += "input file:\n"
            msg += f"{df_missing}"
            warnings.warn(msg)

        # Keep only the columns in cbm_vars_state_df
        cols_to_keep = cbm_vars_state_df.columns.to_list()
        state_updated = state[cols_to_keep].copy()

        # Check state_updated and original cbm_vars_state_df are the same
        # except for the change in growth multiplier i.e. no reordering of rows.
        state_updated_check = state_updated.copy()
        state_updated_check["growth_multiplier"] = state['growth_multiplier_old']
        if not cbm_vars_state_df.equals(state_updated_check):
            msg = "State and state updated do not have the same values. Differences:\n"
            diff = state_updated_check - state_updated
            msg += f"{diff}"
            raise ValueError(msg)

        # Write the state back into cbm_vars
        cbm_vars.state = dataframe.from_pandas(state_updated)
        return cbm_vars



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Constants for eu_cbm_hat package.
Separated to avoid circular imports.
"""

# Built-in modules #
import os
import sys
import pathlib

# First party modules #
from autopaths import Path
from autopaths.dir_path import DirectoryPath
from plumbing.git import GitRepo

# Constants #
project_name = "eu_cbm_hat"
project_url = "https://gitlab.com/bioeconomy/eu_cbm/eu_cbm_hat"
CARBON_FRACTION_OF_BIOMASS = 0.49

# Get paths to module #
self = sys.modules[__name__]
module_dir = Path(os.path.dirname(self.__file__))
module_dir_pathlib = pathlib.Path(os.path.dirname(self.__file__))

# The repository directory #
repos_dir = module_dir.directory

# The module is maybe in a git repository #
git_repo = GitRepo(repos_dir, empty=True)

# Where is the data, default case #
eu_cbm_data_dir = DirectoryPath("~/eu_cbm/eu_cbm_data/")
# But you can override that with an environment variable #
if os.environ.get("EU_CBM_DATA"):
    eu_cbm_data_dir = DirectoryPath(os.environ["EU_CBM_DATA"])
# Prepare the move to pathlib
eu_cbm_data_pathlib = pathlib.Path(str(eu_cbm_data_dir))

# Where are the AIDBs, default case
eu_cbm_aidb_dir = DirectoryPath("~/eu_cbm/eu_cbm_aidb/")
# But you can override that with an environment variable #
if os.environ.get("EU_CBM_AIDB"):
    eu_cbm_aidb_dir = DirectoryPath(os.environ["EU_CBM_AIDB"])
# Prepare the move to pathlib
eu_cbm_aidb_pathlib = pathlib.Path(str(eu_cbm_aidb_dir))





"""Compare substitution scenarios

Note: see the distinction between a scenario combo and a subst_scenario in the
documentation further down below.
"""

import re
from eu_cbm_hat.post_processor.hwp_common_input import hwp_common_input


def compute_substitution(runner, subst_scenario):
    """Substitution scenarios with a reference and a comparison point

     Merge with the data from steel, cement and other materials.

     See the documentation of the compare_substitution function for how to
     compute the difference between the two substitution data frames.

     Example use:

        >>> from eu_cbm_hat.core.continent import continent
        >>> from eu_cbm_hat.post_processor.hwp_substitution  import compute_substitution
        >>> runner = continent.combos['reference'].runners['LU'][-1]
        >>> df = compute_substitution(runner, subst_scenario="reference")

    """
    # Load inflows
    df = runner.post_processor.hwp.build_hwp_stock_since_1990.copy()
    df["sw_inflow"] = df[["sw_con_inflow", "sw_broad_inflow"]].sum(axis=1)
    selected_cols = ["year", "sw_inflow", "wp_inflow", "pp_inflow"]
    df = df[selected_cols]
    # Load split data
    split_wp = hwp_common_input.split_wood_panels.copy()
    # Keep data for the selected country
    selector = split_wp["area"] == runner.country.country_name
    split_wp = split_wp.loc[selector]
    if not len(split_wp) == 1:
        msg = "There should not be more than one value for split_wp\n"
        msg += f"{split_wp}"
        raise ValueError(msg)
    # Split wood panels
    df["wp_fb_inflow"] = df["wp_inflow"] * split_wp["fwp_fibboa"].iloc[0]
    df["wp_pb_inflow"] = df["wp_inflow"] * split_wp["fwp_partboa"].iloc[0]
    # Rename the original inflow columns
    df.rename(columns=lambda x: re.sub(r"inflow", "inflow_0", x), inplace=True)
    # Load substitution parameters
    subst_params = hwp_common_input.subst_params.copy()
    selector = subst_params["subst_scenario"] == subst_scenario
    selector &= subst_params["country"] == runner.country.country_name
    subst_params_ref = subst_params.loc[selector]
    # Merge with substitution parameters
    df = df.merge(subst_params_ref, on="year", how="left")
    # Estimate the avoidance by substitution in wp based substitutes
    cols = subst_params_ref.columns
    frac_cols = cols[cols.str.contains("frac")]
    factor_cols = cols[cols.str.contains("factor")]
    # Check whether all fractions have a corresponding substitution factor
    f_check = [x.replace("frac", "subst_factor") for x in frac_cols]
    missing_factor_cols = set(f_check) - set(factor_cols.to_list())
    if missing_factor_cols:
        msg = "Some fraction columns do not have a corresponding factor column\n"
        msg += f"{missing_factor_cols}"
        raise ValueError(msg)

    # Add fuel wood inflows
    df_fw = runner.post_processor.hwp.ghg_emissions_fw
    col_map = {"tc_primary_fw": "fw_primary_inflow_0",
               "tc_secondary_fw": "fw_secondary_inflow_0"}
    df_fw.rename(columns=col_map, inplace=True)
    df = df.merge(df_fw[["year"] + list(col_map.values())], on="year", how="left")

    # For the substitution we consider only the secondary fuel wood
    for x in ["wp_pb", "wp_fb", "sw", "pp", "fw_primary", "fw_secondary"]:
        # Find which fractions are available for this product
        selected_frac_cols = frac_cols[frac_cols.str.contains(x)].to_list()
        # Create the inflow based on the available fractions and factors
        for frac in selected_frac_cols:
            new_inflow = frac.replace("frac", "inflow")
            factor = frac.replace("frac", "subst_factor")
            df[new_inflow] = df[f"{x}_inflow_0"] * df[frac] * df[factor]

    return df


def compare_substitution(df_ref, df):
    """Compare the substitution data frame to a reference data frame

    1. Compute the difference between the substitution scenario  and the reference
    2. Aggregate and sUm up values

    There is a distinction between forest management and HWP scenarios:

        - a forest management scenario is also called a scenario combination when
          running CBM. There is a combo called "reference".

        - a HWP scenario for example below the subst_scenario arguments are called
          "reference" and "substitution"

    Example compute the difference between two HWP scenarios called "reference"
    and "substitution"within the forest management scenario (i.e. same
    reference combo):

        >>> from eu_cbm_hat.core.continent import continent
        >>> from eu_cbm_hat.post_processor.hwp_substitution  import compare_substitution
        >>> from eu_cbm_hat.post_processor.hwp_substitution  import compute_substitution
        >>> runner = continent.combos['reference'].runners['LU'][-1]
        >>> df_ref = compute_substitution(runner, subst_scenario="reference")
        >>> df_subst = compute_substitution(runner, subst_scenario="substitution")
        >>> # Comparison
        >>> compare_substitution(df_ref, df_subst)

    """
    # Select only the inflow columns
    selector = df_ref.columns.str.contains("inflow")
    # Except the original inflow columns
    selector &= ~df_ref.columns.str.contains("inflow_0")
    inflow_cols = df_ref.columns[selector].to_list()
    df_diff = df_ref[["year"] + inflow_cols].copy()
    # Compute the difference between inflows
    df_diff[inflow_cols] = df[inflow_cols] - df_ref[inflow_cols]
    # Rename to savings
    new_name = "savings"
    df_diff.rename(columns=lambda x: re.sub(r"inflow", new_name, x), inplace=True)
    # Sum up all the above per year
    cols = df_diff.columns[df_diff.columns.str.contains(new_name)]
    df_diff["total_savings"] = df_diff[cols].sum(axis=1)
    # Convert to CO2
    df_diff["total_savings_CO2"] = df_diff["total_savings"] * 44 / 12
    return df_diff







import os
import shutil
import pandas

from plumbing.logger import create_file_logger

from eu_cbm_hat import eu_cbm_data_dir
from eu_cbm_hat.crcf import Runner
from eu_cbm_hat.crcf.driver_netcdf import Netcdf
from eu_cbm_hat.crcf.driver_database import Database




class CountrySimulator:
    """
    Extract all unique stand-event combinations from the eu-cbm-data dataset.
    Store these combinations in a database file.
    Run simulations for each individual stand and save the results in a NetCDF file.

    How to run:
        >>> from eu_cbm_hat.crcf import CountrySimulator
        >>> country_code = 'FI'
        >>> output_dir   = 'results'
        >>> obj = CountrySimulator()
        >>> obj.run(country_code, output_dir)
    """

    classifiers = 'status,forest_type,region,mgmt_type,mgmt_strategy,climate,con_broad,site_index,growth_period'.split(',')
    columns_events = 'status,forest_type,region,mgmt_type,mgmt_strategy,climate,con_broad,site_index,growth_period,using_id,sw_start,sw_end,hw_start,hw_end,min_since_last_dist,max_since_last_dist,last_dist_id,min_tot_biom_c,max_tot_biom_c,min_merch_soft_biom_c,max_merch_soft_biom_c,min_merch_hard_biom_c,max_merch_hard_biom_c,min_tot_stem_snag_c,max_tot_stem_snag_c,min_tot_soft_stem_snag_c,max_tot_soft_stem_snag_c,min_tot_hard_stem_snag_c,max_tot_hard_stem_snag_c,min_tot_merch_stem_snag_c,max_tot_merch_stem_snag_c,min_tot_merch_soft_stem_snag_c,max_tot_merch_soft_stem_snag_c,min_tot_merch_hard_stem_snag_c,max_tot_merch_hard_stem_snag_c,efficiency,sort_type,measurement_type,amount,dist_type_name,step'.split(',')

    def __init__(self,
                 path_cache:str='cache',
                 scenario:str='reference',
                 num_timesteps:int=200,
                ):

        self.cache = path_cache
        self.scenario = scenario
        self.num_timesteps = num_timesteps

        self.log = None
        self.country = None


    def init_libcbm_input_files(self):

        # copy default files
        shutil.copyfile(f'{eu_cbm_data_dir}/countries/{self.country}/common/age_classes.csv', f'{self.cache}/input/csv/age_classes.csv')
        shutil.copyfile(f'{eu_cbm_data_dir}/countries/{self.country}/common/classifiers.csv', f'{self.cache}/input/csv/classifiers.csv')
        shutil.copyfile(f'{eu_cbm_data_dir}/countries/{self.country}/common/disturbance_types.csv', f'{self.cache}/input/csv/disturbance_types.csv')

        # growth_curves.csv
        self.write_growth_curves()

        # transitions.csv (empty file)
        with open(f'{eu_cbm_data_dir}/countries/{self.country}/activities/mgmt/transitions.csv') as ff:
            header = ff.readline()
        with open(f'{self.cache}/input/csv/transitions.csv', 'w') as ff:
            ff.write(header)


    def write_growth_curves(self):
        ds = pandas.read_csv(f'{eu_cbm_data_dir}/countries/{self.country}/activities/mgmt/growth_curves.csv', dtype=str)
        ds = ds[ds['scenario'] == self.scenario].drop(columns=['scenario'])
        ds.to_csv(f'{self.cache}/input/csv/growth_curves.csv', index=False)


    def read_growth_curves(self)->pandas.DataFrame:
        ds = pandas.read_csv(f'{eu_cbm_data_dir}/countries/{self.country}/activities/mgmt/growth_curves.csv', dtype=str)
        ds = ds[ds['scenario'] == self.scenario].drop(columns=['scenario'])
        unique = ds.drop_duplicates(subset=self.classifiers)
        return unique


    def read_classifiers(self)->dict:

        # read eu_cbm_hat classifiers.csv
        df0 = pandas.read_csv(f'{eu_cbm_data_dir}/countries/{self.country}/common/classifiers.csv')

        # data to dict
        classifiers = {}
        for ii in df0['classifier_number'].unique():
            df1  = df0[df0['classifier_number']==ii]
            name = df1.loc[df1['classifier_value_id']=='_CLASSIFIER', 'name'].item()
            classifiers[name] = {}
            for _, hh in df1.loc[df1['classifier_value_id']!='_CLASSIFIER', ['classifier_value_id', 'name']].iterrows():
                classifiers[name][hh['classifier_value_id']] = hh['name']

        return classifiers


    def read_disturbance_types(self)->dict:
        ds = pandas.read_csv(f'{eu_cbm_data_dir}/countries/{self.country}/common/disturbance_types.csv', dtype=str)
        ds = ds.set_index(ds.columns[0])[ds.columns[1]].to_dict()
        ds = {'-1': 'none'} | ds
        return ds


    def read_events(self)->pandas.DataFrame:
        ds = pandas.read_csv(f'{eu_cbm_data_dir}/countries/{self.country}/activities/mgmt/events.csv', dtype=str)
        ds = ds[ds['scenario'] == self.scenario].drop(columns=['scenario'])
        return ds


    def read_inventory(self)->pandas.DataFrame:

        ds = pandas.read_csv(f'{eu_cbm_data_dir}/countries/{self.country}/activities/mgmt/inventory.csv', dtype=str)
        ds = ds[ds['scenario'] == self.scenario].drop(columns=['scenario'])
        ds = ds.iloc[ds[self.classifiers].drop_duplicates().index]

        ds.loc[:, 'using_id'] = False
        ds.loc[:, 'age'] = 1
        ds.loc[:, 'area'] = 100.0

        return ds


    def write_empty_events(self):
        with open(f'{self.cache}/input/csv/events.csv', 'w') as ff:
            ff.write(','.join(self.columns_events) + '\n')


    def allowed_disturbances(self):

        # read eu_cbm_hat disturbance_types.csv
        dist = self.read_disturbance_types()

        # False for events involving fire, wind, insects
        allowed = {}
        for key, value in dist.items():
            value = value.lower()
            if 'fire' in value or 'wind' in value or 'insect' in value:
                allowed[key] = False
            else:
                allowed[key] = True

        return allowed


    def find_corresponding_rows(self, df_to_search:pandas.DataFrame, df_reference:pandas.Series)->pandas.DataFrame:

        # replace '?' with df_reference corresponding values
        tmp = df_to_search.copy()
        for label in self.classifiers:
            tmp.loc[tmp[label]=='?', label] = df_reference[label]

        # rows in df_to_search corresponding to df_reference[self.classifiers]
        ds = tmp[tmp[self.classifiers[:-1]].eq(df_reference[self.classifiers[:-1]]).all(axis=1)].reset_index()

        return ds


    def find_init_curr_curves_indices(self, curves:pandas.DataFrame, stand:pandas.Series)->tuple:

        # corresponding rows
        df = self.find_corresponding_rows(curves, stand)
        if len(df) > 2:
            self.log.error(f'found more than 2 corresponding growth curves\n\n{df}')
            raise ValueError(f'found more than 2 corresponding growth curves\n\n{df}')

        # cur and init curves indices
        ds0 = df.loc[df['growth_period'] == 'Init', 'index']
        ds1 = df.loc[df['growth_period'] == 'Cur', 'index']
        index_init = -1 if len(ds0) == 0 else ds0.item()
        index_curr = -1 if len(ds1) == 0 else ds1.item()

        return index_init, index_curr


    def run(self, country_code:str, output_dir:str, dry_run:bool=False):

        # set country code
        self.country = country_code

        # logger
        logger_name = country_code
        log_file    = f'{output_dir}/simulations_{country_code}.log'
        self.log    = create_file_logger(logger_name, log_file, console_level='debug')

        # retrieve eu-cbm-hat info
        curves = self.read_growth_curves()
        stands = self.read_inventory()
        events = self.read_events()
        dist   = self.allowed_disturbances()

        # init database
        path_db = f'sqlite:///{output_dir}/simulations_{self.country}.db'
        db = Database(path_db)
        db.init(self.read_classifiers(), self.read_disturbance_types())

        # init env for eu-cbm-hat computations
        if not dry_run:

            # create cache folder
            os.makedirs(f'{self.cache}/input/csv', exist_ok=True)

            # init libcbm input files
            self.init_libcbm_input_files()

            # netcdf instance
            nc = Netcdf(f'{output_dir}/simulations_{self.country}.nc')

        # avoid duplicate eu-cbm-hat simulation runs.
        # sidx = list[idx_sim], where idx_sim =
        # (idx row growth curve init, idx row growth curve curr, idx row event)
        sidx = []

        # loop on species
        for _, stand in stands.iterrows():

            # =====================
            #     simple growth
            # =====================

            # log
            self.log.info('==========')
            self.log.info(f'eu_cbm_hat on {",".join([str(ii) for ii in stand[self.classifiers[:-1]].to_list()])}')
            self.log.info('==========')

            # add species to db
            id_species = db.add_commit_species(stand)

            # assemble idx_sim
            idx_curve = self.find_init_curr_curves_indices(curves, stand)
            idx_sim   = idx_curve + (-1,)

            if idx_sim not in sidx:

                # add simulation to db
                db.add_simulation(id_species, '-1', -1, len(sidx))
                sidx.append(idx_sim)

                # cbm simulation
                if not dry_run:

                    # write libcbm inventory.csv and events.csv files
                    stand.to_frame().T.to_csv(f'{self.cache}/input/csv/inventory.csv', index=False)
                    self.write_empty_events()

                    # run cbm
                    table = Runner(self.cache, self.country, self.num_timesteps, logger_name).run()
                    for classif in self.classifiers:
                        table[classif] = table[classif].replace(db.classifier_ids(classif))

                    # eventually init netcdf
                    if not nc.is_init:
                        nc.init(self.num_timesteps+1, table.columns)

                    # write to netcdf
                    nc.write_simulation(table.to_numpy())
            
            else:

                # add simulation to db
                db.add_simulation(id_species, '-1', -1, sidx.index(idx_sim))
                self.log.info('Previously computed')


            # ============
            #     mgmt
            # ============

            # unique events
            ds = self.find_corresponding_rows(events, stand)
            oo = 'sw_start,hw_start,dist_type_name'.split(',')
            ds = ds.iloc[ds[oo].drop_duplicates().index]

            # loop on events
            for _, evt in ds.iterrows():

                # name and timestep
                evt_name = evt['dist_type_name']
                evt_step = '1' if evt["sw_start"] == '0' else evt["sw_start"]

                # eventually skip if not allowed disturbance
                if not dist[evt_name]:
                    continue

                # log
                self.log.info('==========')
                self.log.info(f'eu_cbm_hat on {",".join([str(ii) for ii in stand[self.classifiers[:-1]].to_list()])},{evt_name},{evt_step}')
                self.log.info('==========')

                # assemble idx_sim
                idx_sim = idx_curve + (evt['index'],)

                if idx_sim not in sidx:

                    # add simulation to db
                    db.add_simulation(id_species, evt_name, evt_step, len(sidx))
                    sidx.append(idx_sim)

                    # cbm simulation
                    if not dry_run:

                        # build data for events.csv
                        # add [sort_type, measurement_type, amount, evt_name, evt_step]
                        data  = evt[self.columns_events[:-5]].to_list()
                        data += [1, 'P', 1, evt_name, evt_step]

                        # write events.csv
                        df = pandas.DataFrame([data], columns=self.columns_events)
                        df.to_csv(f'{self.cache}/input/csv/events.csv', index=False)

                        # run cbm
                        table = Runner(self.cache, self.country, self.num_timesteps, logger_name).run()
                        for classif in self.classifiers:
                            table[classif] = table[classif].replace(db.classifier_ids(classif))

                        # write to netcdf
                        nc.write_simulation(table.to_numpy())

                else:

                    # add simulation to db
                    db.add_simulation(id_species, evt_name, evt_step, sidx.index(idx_sim))
                    self.log.info('Previously computed')

            # commit
            db.commit()

        # close db
        db.close()

        # clean cache
        if os.path.exists(self.cache):
            shutil.rmtree(self.cache)

        # log
        self.log.info('Normal termination')


"""
The purpose of this script is to compute the NPP coresponding to cbm's runs
"""

from functools import cached_property
from typing import Union, List
import warnings
import pandas
import numpy as np

def npp_components(df: pandas.DataFrame, groupby: Union[List[str], str]):
    """Compute the Net Annual Increment and Gross Annual Increment on climates, for npp-nai analysis

    Based on stock change and movements to the product pools as well as
    turnover and mouvements to air.

   
        >>> from eu_cbm_hat.core.continent import continent
        >>> from eu_cbm_hat.post_processor.nai import NAI_AGG_COLS
        >>> from eu_cbm_hat.post_processor.nai import compute_nai_gai
        >>> runner = continent.combos['reference'].runners['ZZ'][-1]
        >>> index = ["status"]
        >>> nai_st = runner.post_processor.nai.df_agg(index)


    """

    print(groupby)

    
    if isinstance(groupby, str):
        groupby = [groupby]

    if "year" in groupby:
        msg = " This functions computes the difference in stock across groups "
        msg += "through time so 'year' should not be in the group variables:\n"
        msg += f"{groupby}"
        raise ValueError(msg)

    # Order by groupby variables, then years
    df.sort_values(groupby + ["year"], inplace=True)

    #print(df.head(2))

    
    # Check that there are no duplications over the groupby variables plus year
    #selector = df[["year"] + groupby].duplicated(keep=False)
    #if any(selector):
    #    msg = "The following rows have duplications along the groupby variables.\n"
    #    msg += f"{df.loc[selector, ['year'] + groupby ]}"
    #    msg += "\nPlease aggregate first along the groupby variables and year:\n"
    #    msg += f"{['year'] + groupby }\n Then run this function.\n"
    #    raise ValueError(msg)

    # Compute the difference in stock for the total standing biomass
    # Use Observed = True to avoid the warning when using categorical variables
    df["agb"] = df["merch"] + df["other"]+ df["foliage"]+ df["roots"]
    df["net_agb"] = df.groupby(groupby, observed=True)["agb"].diff()
   
    # Compute NAI for the merchantable pool
    df["npp"] = df[["net_agb",
                    # biomass to production
                    "merch_prod", 
                    "oth_prod",
                    # transfers to non living pools through disturbances
                    'disturbance_merch_litter_input',
                    'disturbance_oth_litter_input',
                    'disturbance_fol_litter_input',
                    'disturbance_coarse_litter_input',
                    'disturbance_fine_litter_input',        
                    'disturbance_merch_to_air',
                    'disturbance_oth_to_air',
                    'disturbance_fol_to_air',
                    'disturbance_coarse_to_air',
                    'disturbance_fine_to_air',
                    # transfers to non living pools through natural decay
                    'turnover_merch_litter_input',
                    'turnover_oth_litter_input',
                    'turnover_fol_litter_input',
                    'turnover_coarse_litter_input',
                    'turnover_fine_litter_input']].sum(axis=1)

    # Compute per hectare values
    #df["nai_merch_ha"] = df["nai_merch"] / df["area"]
    #df["gai_merch_ha"] = df["gai_merch"] / df["area"]
    #df["nai_agb_ha"] = df["nai_agb"] / df["area"]
    #df["gai_agb_ha"] = df["gai_agb"] / df["area"]
    return df


class NPP:
    """Compute the net growth
    Usage:

        >>> # Net Annual Increment of the merchantable pool (nai_merch) and of
        >>> # all the above ground biomass (nai_agb) by status
        >>> nai_st = runner.post_processor.nai.df_agg(["status"])
        >>> selector = nai_st["status"] == 'ForAWS'
        >>> nai_st.loc[selector, ["year", "status", "area", "nai_merch", "nai_agb"]].head()
     
        """

    def __init__(self, parent):
        self.parent = parent
        self.runner = parent.runner
        self.combo_name = self.runner.combo.short_name

    @cached_property
    def compute_npp(self):
        """Merchantable pools and fluxes aggregated at the classifiers level"""
        df = self.parent.pools_fluxes_morf
        print(df.columns)
        df_out = npp_components(df, groupby= ['status','climate', 'con_broad'])
        return df_out



"""Input data for a small self contained runner-type of object called a bud.
"""
import pandas as pd
from autopaths.auto_paths import AutoPaths


class BudInputData:
    """
    CSV files from a scenario "input/csv" directory used as input to CBM

    Examples
    --------
    >>> input_data = InputData("scenarios/reference/input/csv/")
    >>> df = input_data["inventory"]  # Loads inventory.csv as DataFrame
    """

    all_paths = """
    /input/csv/
    /input/csv/age_classes.csv
    /input/csv/classifiers.csv
    /input/csv/disturbance_types.csv
    /input/csv/events.csv
    /input/csv/inventory.csv
    /input/csv/transitions.csv
    /input/csv/growth_curves.csv
    """

    def __init__(self, parent):
        """
        Initialize InputData with the path to the CSV directory.

        Parameters
        ----------
        parent: bud object which has a parameter do the data directory,
            where a sub-path contains the directory containing CSV files.
        """
        self.parent = parent
        # pathlib.Path object (we would prefer to use pathlib)
        self.csv_directory = self.parent.data_dir / "input/csv"
        # AutoPaths object (for compatibility with runner.input_data)
        self.paths = AutoPaths(str(self.parent.data_dir), self.all_paths)
        # Verify the CSV directory exists
        if not self.csv_directory.exists():
            raise FileNotFoundError(f"Directory not found: {self.csv_directory}")

        if not self.csv_directory.is_dir():
            raise NotADirectoryError(f"Path is not a directory: {self.csv_directory}")

    def __getitem__(self, key: str) -> pd.DataFrame:
        """
        Load and return a CSV file as a pandas DataFrame.

        Parameters
        ----------
        key : str
            Name of the CSV file (without .csv extension).

        Returns
        -------
        pandas.DataFrame
            The loaded CSV data.

        Raises
        ------
        FileNotFoundError
            If the CSV file doesn't exist.
        Exception
            If there's an error reading the CSV file.
        """
        csv_file = self.csv_directory / f"{key}.csv"
        if not csv_file.exists():
            available_files = [f.stem for f in self.csv_directory.glob("*.csv")]
            raise FileNotFoundError(
                f"CSV file '{key}.csv' not found in {self.csv_directory}. "
                f"Available files: {available_files}"
            )
        try:
            df = pd.read_csv(csv_file)
            return df

        except Exception as e:
            raise Exception(f"Error reading CSV file '{csv_file}': {str(e)}")

    def __contains__(self, key: str) -> bool:
        """
        Check if a CSV file exists in the directory.

        Parameters
        ----------
        key : str
            Name of the CSV file (without .csv extension).

        Returns
        -------
        bool
            True if the file exists, False otherwise.
        """
        csv_file = self.csv_directory / f"{key}.csv"
        return csv_file.exists()

    def list_available_files(self) -> list:
        """
        List all available CSV files in the directory.

        Returns
        -------
        list of str
            List of available CSV file names (without .csv extension).
        """
        return [f.stem for f in self.csv_directory.glob("*.csv")]

    def get_file_info(self, key: str) -> dict:
        """
        Get information about a CSV file.

        Parameters
        ----------
        key : str
            Name of the CSV file (without .csv extension).

        Returns
        -------
        dict
            Information about the file including size, modification time, etc.

        Raises
        ------
        FileNotFoundError
            If the CSV file doesn't exist.
        """
        csv_file = self.csv_directory / f"{key}.csv"

        if not csv_file.exists():
            raise FileNotFoundError(f"CSV file '{key}.csv' not found")

        stat = csv_file.stat()
        return {
            "file_path": str(csv_file),
            "size_bytes": stat.st_size,
            "modified_time": stat.st_mtime,
        }



import logging

from autopaths.auto_paths import AutoPaths

from plumbing.cache  import property_cached
from plumbing.common import camel_to_snake
from plumbing.logger import create_file_logger

from eu_cbm_hat                     import eu_cbm_data_dir
from eu_cbm_hat.cbm.simulation      import Simulation
from eu_cbm_hat.launch.create_json  import CreateJSON
from eu_cbm_hat.launch.associations import Associations
from eu_cbm_hat.info.orig_data      import OrigData

# all_country_codes = pandas.read_csv(eu_cbm_data_dir + 'common/country_codes.csv')
# ref_years = pandas.read_csv(eu_cbm_data_dir + 'common/reference_years.csv')


class AIDB:
    all_paths = '/config/aidb.db'
    def __init__(self, data_dir):
        self.paths = AutoPaths(data_dir, self.all_paths)

class InputData:
    all_paths = """
    /input/csv/
    /input/csv/age_classes.csv
    /input/csv/classifiers.csv
    /input/csv/disturbance_types.csv
    /input/csv/events.csv
    /input/csv/inventory.csv
    /input/csv/transitions.csv
    /input/csv/growth_curves.csv
    """
    def __init__(self, input_dir):
        self.paths = AutoPaths(input_dir, self.all_paths)

class Country:
    def __init__(self, iso2_code):
        self.iso2_code = iso2_code
        self.data_dir  = f'{eu_cbm_data_dir}/countries/{self.iso2_code}'
        self.aidb      = AIDB(self.data_dir)
    @property_cached
    def associations(self):
        return Associations(self)
    @property_cached
    def orig_data(self):
        return OrigData(self)

class Output:
    def __init__(self, parent):
        # classifiers
        classifiers_ids = parent.simulation.sit.classifier_value_ids
        classifiers     = parent.simulation.cbm_output.classifiers.to_pandas()
        value2id        = {v: k for m in classifiers_ids.values() for k, v in m.items()}
        classifiers     = classifiers.set_index(['identifier', 'timestep'])
        classifiers     = classifiers.replace(value2id)
        classifiers     = classifiers.reset_index()
        # data
        area            = parent.simulation.cbm_output.area.to_pandas()
        flux            = parent.simulation.cbm_output.flux.to_pandas()
        parameters      = parent.simulation.cbm_output.parameters.to_pandas()
        pool            = parent.simulation.cbm_output.pools.to_pandas()
        state           = parent.simulation.cbm_output.state.to_pandas()
        on_cols         = ['identifier', 'timestep']
        self.table      = (classifiers
                            .merge(area,       how='left', on=on_cols)
                            .merge(parameters, how='left', on=on_cols)
                            .merge(state,      how='left', on=on_cols)
                            .merge(pool,       how='left', on=on_cols)
                            .merge(flux,       how='left', on=on_cols)
                          )
        self.table.columns = self.table.columns.to_series().apply(camel_to_snake)
    def save(self, filename):
        self.table.to_csv(filename, index=False)

class Runner:
    all_paths = """
    /input/
    /input/json/config.json
    /runner.log
    /cbm_output.csv
    """
    def __init__(self, path, country_iso2_code, num_timesteps=1, logger=None):
        self.short_name    = country_iso2_code
        self.num_timesteps = num_timesteps
        self.country       = Country(country_iso2_code)
        self.input_data    = InputData(path)
        self.paths         = AutoPaths(path, self.all_paths)
        self.log           = create_file_logger(self.country.iso2_code, self.paths.log, console_level='debug') if logger is None else logging.getLogger(logger)
    @property_cached
    def create_json(self):
        return CreateJSON(self)
    @property_cached
    def simulation(self):
        return Simulation(self)
    @property_cached
    def cbm_output(self):
        return Output(self)
    def run(self, output:str=None, interrupt_on_error=True):
        self.create_json()
        self.log.info("Running simulation")
        self.simulation(interrupt_on_error)
        self.log.info("Simulation finished")
        if output is not None:
            self.log.info(f"Writing output to {output}")
            self.cbm_output.save(output)
        else:
            return self.cbm_output.table


"""Adapt the age_classes.csv input file to the growth cuves"""

import pandas as pd

def adapt_age_classes(growth_curves, age_class_length):
    """Adapt age_classes to the max age given in the growth curve and to the age class length

    Reduce the length of the age_classes.csv file according to the number of
    columns supplied in the growth_curves.csv in order to avoid this error:
    ValueError: expected 41 columns. This is defined as 10 classifiers plus 31 age classes
    File ~/repos/eu_cbm/libcbm_py/libcbm/input/sit/sit_yield_parser.py:54, in
    parse(yield_table, classifiers, classifier_values, age_classes)

    Example:

        >>> import pandas as pd
        >>> from eu_cbm_hat.bud.adapt_age_classes import adapt_age_classes
        >>> growth_curves = pd.DataFrame({"vol0" : [0], "vol1": [30], "vol2": [90]})
        >>> adapt_age_classes(growth_curves, 10)
        >>> adapt_age_classes(growth_curves, 1)

    """
    if age_class_length not in [1, 10]:
        msg = "Age class length can only take a value of 1 or 10."
        msg += f"\nCurrently age_class_length: {age_class_length}"
        raise ValueError(msg)
    vol_cols = growth_curves.columns[growth_curves.columns.str.contains("^vol")].to_list()
    df = pd.DataFrame({"vol_col" : vol_cols})
    df["age_class_id"] = df["vol_col"].str.replace("vol", "AGEID")
    df["size"] = age_class_length
    df.loc[0, "size"] = 0
    df.drop(columns={"vol_col"}, inplace=True)
    return df



""" Common input file used for climate adjustment of growth based on modelled NPP values

Written by Viorel Blujdea and Paul Rougieux.

JRC Biomass Project. Unit D1 Bioeconomy.

- See also plots of NPP in `eu_cbm_hat.plot.npp`:

    >>> import matplotlib.pyplot as plt
    >>> from eu_cbm_hat.plot.npp import plot_npp_facet
    >>> from eu_cbm_hat.info.clim_adjust_common_input import mean_npp_by_model_country_clu_con_broad
    >>> df = mean_npp_by_model_country_clu_con_broad(hist_start_year=2010, hist_end_year=2020)
    >>> plot_npp_facet(df, 'Austria')
    >>> plt.show()

"""

import pandas as pd
from eu_cbm_hat.constants import eu_cbm_data_pathlib


def mean_npp_by_model_country_clu_con_broad(hist_start_year, hist_end_year):
    """Read common input file mean NPP by model country CLU and con_broad


    The growth curves is based on a NAI value from the NFI which already
    includes the impact of droughts or other events so we cannot modify it too
    much. For the future, we need to capture both extreme values and the trend.

    A given stand can only have one growth curve calibrated over the historical
    period. We therefore need our growth modifier value to have an average
    value of 1 over the historical period. We compute the average historical
    NPP over the period for which the growth curve is valid. For example, if
    our reference period is 2010-2020. That means we take the average NPP over
    2010-2020 and we use this as the denominator to compute a NPP ratio. Then
    we divide each years's NPP through the average to obtain the growth
    modifier value.

    Usage:

        >>> from eu_cbm_hat.info.clim_adjust_common_input import mean_npp_by_model_country_clu_con_broad
        >>> df = mean_npp_by_model_country_clu_con_broad(hist_start_year=2010, hist_end_year=2020)

    """
    csv_filename = "mean_npp_by_model_country_clu_con_broad.csv"
    df = pd.read_csv(eu_cbm_data_pathlib / "common" / csv_filename)

    # REMOVE renaming dut to change of column names in the input file. Workflow from TRENDY inputs is a csv chnaged to
    # 'mean_npp_by_model_country_clu_con_broad_original.csv'. The preprocesing with explore/users_tools/npp_input_processing.py adds
    # a 'mix-models' scenario,  a kind of average of all-models
    #col_rename = {
    #       "npp (tC/ha/yr)": "npp",
    #       "forest_type": "con_broad",
    #       "climatic_unit": "climate",
    #   }
    # Check for missing columns before renaming
    #missing_col = set(col_rename.keys()) - set(df.columns)
    #if missing_col: 
    #    msg = "The following columns are supposed to be renamed "
    #    msg += f"but they are missing: {missing_col}"
    #    msg += "\nData frame columns:\n"
    #    msg += f"{df.columns}"
    #    raise ValueError(msg)
    #df.rename(columns=col_rename, inplace=True)
    
    # Convert climate to a character variable for compatibility with CBM classifiers
    df["climate"] = df["climate"].astype(str)
    # Group the data by 'model', 'country', 'forest_type', and 'climatic_unit'
    # and calculate the first year's 'npp' value for each group
    index = ["model", "country", "con_broad", "climate"]
    # Compute the average over the historical period
    selector = df["year"] >= hist_start_year
    selector &= df["year"] <= hist_end_year
    df_hist_mean = (df.loc[selector].groupby(index).agg(hist_mean_npp = ("npp", "mean"))).reset_index()
    # Merge the first year's 'npp' values with the original DataFrame
    df = df.merge(df_hist_mean, on=index)
    # Calculate the ratio of each year's 'npp' value to historical mean npp
    df["ratio"] = df["npp"] / df["hist_mean_npp"]
    # Rename con broad
    df["con_broad"] = df["con_broad"].replace({"BL": "broad", "NL": "con"})
    # "default" is a reserved value for the case where there is no climate adjustment
    selector = df["model"] == "default"
    if any(selector):
        msg = "'default' is not allowed as a model name. "
        msg += "It is reserved for the case where no climate model is used\n"
        msg += f"{df.loc[selector]}"
        raise ValueError(msg)
    return df



"""Functions to select HWP scenario objects part of the post processor

- Select a harvest wood product (HWP) scenario along with options for
  recycling, and trade assumptions. Returns a configured `post_processor.hwp`
  object for downstream simulation and result comparison.

- The substitution is performed in an post_processor/hwp_substitution.py An
  extra function that is independent of the post_processor hwp object and returns
  a data frame for comparison purposes.
     - def compute_substitution(runner, subst_scenario)
     - def compare_substitution(df_ref, df)


"""
from typing import Optional, Union
from eu_cbm_hat.core.continent import continent

def select_hwp_scenario(
    iso2_code: str,
    forest_mgmt_scenario: str,
    hwp_frac_scenario: str,
    n_years_dom_frac: int,
    no_export_no_import: bool,
    recycling: bool = True,
):
    """
    Select a harvest wood product (HWP) scenario along with options for recycling,
    and trade assumptions. Returns a configured `post_processor.hwp`
    object for downstream simulation and result comparison.

    Substitution should be run by compare_substitution in a next step.

    Parameters
    ----------
    iso2_code : str
        Two-letter ISO2 country code (e.g., "LU" for Luxembourg).
    forest_mgmt_scenario : str, optional
        Name of the forest management scenario (e.g., "reference", "intensive",
        etc.). It is a combo name used to select the corresponding runner for
        the given scenario combination runner in `continent.combos`.
    hwp_frac_scenario : str, optional
        Name of the HWP fraction scenario (e.g., "default", "more_sw").
        Defines allocation rules for harvested wood products.
    n_years_dom_frac : int, optional
        Number of years of historical domestic fraction data to be used for
        estimating HWP flow shares.
    no_export_no_import : bool, optional
        If True, sets all import/export correction factors to unity, effectively
        disabling cross-border trade effects in HWP estimations.
    recycling : bool, default=True
        If True, includes recycling processes in the HWP model.
        Representation may depend on the underlying implementation.

    Returns
    -------
    hwp : post_processor.hwp
        A configured HWP post-processor object with the chosen scenario settings.

    Examples
    --------
    Select two different HWP scenarios and compare outputs:

    >>> from eu_cbm_hat.post_processor.select_hwp_scenario import select_hwp_scenario
    >>> # Reference scenario
    >>> hwp_ref_frac = select_hwp_scenario(
    ...     iso2_code="LU",
    ...     forest_mgmt_scenario="reference",
    ...     hwp_frac_scenario="default",
    ...     n_years_dom_frac=3,
    ...     no_export_no_import=False,
    ...     recycling=True,
    ... )
    >>> # Scenario with more sawnwood allocation
    >>> hwp_more_sw = select_hwp_scenario(
    ...     iso2_code="LU",
    ...     forest_mgmt_scenario="reference",
    ...     hwp_frac_scenario="more_sw",
    ...     n_years_dom_frac=3,
    ...     no_export_no_import=False,
    ...     recycling=True,
    ... )
    >>> # Compare intermediate tables
    >>> print(hwp_ref_frac.prod_from_dom_harv_sim)
    >>> print(hwp_more_sw.prod_from_dom_harv_sim)
    >>> # Compare HWP sink results
    >>> print(hwp_ref_frac.stock_sink_results)
    >>> print(hwp_more_sw.stock_sink_results)

    In a further step you can compare 2 substitution scenarios with
    `compare_substitution`().

    """
    # Select the post processor HWP object
    hwp = continent.combos[forest_mgmt_scenario].runners[iso2_code][-1].post_processor.hwp
    # Define properties
    hwp.hwp_frac_scenario = hwp_frac_scenario
    hwp.n_years_dom_frac = n_years_dom_frac
    # Add recycling information or not
    hwp.add_recycling = recycling
    # Set export import factors to one
    hwp.no_export_no_import = no_export_no_import
    return hwp


def stock_sink_results(**kwargs):
    """Result data frame for the given scenario combination and HWP scenario.

    Return the output data frame of the stock_sink_results method. Add the name
    of all scenarios as a column name.

    For example compare results tables:

        >>> from eu_cbm_hat.post_processor.select_hwp_scenario import stock_sink_results
        >>> hwp_refd = stock_sink_results(iso2_code="LU", combo="reference", hwp_frac="default")
        >>> hwp_more_sw = stock_sink_results(iso2_code="LU", combo="reference", hwp_frac="more_sw")

    """
    hwp = select_hwp_scenario(**kwargs)
    df = hwp.stock_sink_results
    df["forest_mgmt_scenario"] = hwp.runner.combo.short_name
    # Place the last column first
    cols = df.columns.to_list()
    cols = cols[-1:] + cols[:-1]
    df = df[cols]
    return df






""" The purpose of this script is to copy data necessary to test the bud object.

Usage:

    ipython -i ~/eu_cbm/eu_cbm_hat/scripts/setup/copy_zz_from_eu_cbm_data_to_bud_test_dir.py

It has to be run on a machine that has the eu_cbm_data repository and the
output of a ZZ model run. i.e. run ZZ first with:

     ipython3 -i -- ~/repos/eu_cbm/eu_cbm_hat/scripts/running/run_zz.py

"""

import shutil
from eu_cbm_hat import eu_cbm_data_pathlib
from eu_cbm_hat import module_dir_pathlib

# Path to copy data from
zz_associations_path = eu_cbm_data_pathlib / "countries/ZZ/config/associations.csv"
zz_reference_output_dir = eu_cbm_data_pathlib / "output/reference/ZZ/0"
if not zz_reference_output_dir.exists():
    msg = "Run the model first with:\n"
    msg += "ipython -i -- ~/repos/eu_cbm/eu_cbm_hat/scripts/running/run_zz.py"
    raise ValueError(msg)

# Path to copy data to
test_data_dir = module_dir_pathlib / "tests/bud_data"
if not test_data_dir.exists():
    test_data_dir.mkdir(parents=True)

associations_path = test_data_dir / "input/csv/associations.csv"
print(f"\nCopy {zz_associations_path}")
print(f"to {associations_path}")
shutil.copy(zz_associations_path, associations_path)

# libcbm input file paths to copy data from
# I copied the output of the bash command
# cd ~/eu_cbm/eu_cbm_data/output/reference/ZZ/0/
# find input -type f
files = """input/csv/inventory.csv
input/csv/disturbance_types.csv
input/csv/growth_curves.csv
input/csv/transitions.csv
input/csv/classifiers.csv
input/csv/events.csv
input/csv/age_classes.csv
"""

print(f"\nCopy\n{files}")
print(f"from \n {zz_reference_output_dir}\nto\n {test_data_dir}")
for line in files.splitlines():
    orig_file = zz_reference_output_dir / line
    dest_file = test_data_dir / line
    dest_file.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(orig_file, dest_file)

